# -*- coding: utf-8 -*-
"""
Base unparser implementation

Brings together the different bits from the different helpers.
"""

from __future__ import unicode_literals

import logging

from calmjs.parse.unparsers.walker import (
    Dispatcher,
    walk,
)
from calmjs.parse.handlers.core import default_rules
from calmjs.parse.handlers.core import token_handler_str_default

logger = logging.getLogger(__name__)


class BaseUnparser(object):
    """
    A simple base class for gluing together the default Dispatcher and
    walk function together to achieve unparsing.
    """

    def __init__(
            self,
            definitions,
            token_handler=None,
            rules=(default_rules,),
            layout_handlers=None,
            deferrable_handlers=None,
            prewalk_hooks=(),
            walk=walk,
            dispatcher_cls=Dispatcher):
        """
        Optional arguements

        definition
            The definition for unparsing.
        token_handler
            passed onto the dispatcher object; this is the handler that
            will process the token in to chunks.
        rules
            A tuple of callables that will set up the various rules that
            will be passed to the dispatcher instance.  It should return
            the mappings for layout_handlers and deferrable_handlers.
        layout_handlers
            Additional layout handlers for the Dispatcher instance.
        deferrable_handlers
            Additional deferrable handlers for the Dispatcher instance.
        prewalk_hooks
            A list of callables that will be called before the walk
            function is called.  The dispatcher instance and the node
            that was called on will be passed to these callables.
        walk
            The walk function - defaults to the version from the walker
            module
        dispatcher_cls
            The Dispatcher class - defaults to the version from the
            walker module
        """

        # the base items.
        self.definitions = {}
        self.definitions.update(definitions)
        self.walk = walk
        self.dispatcher_cls = dispatcher_cls

        self.rules = rules

        self.layout_handlers = layout_handlers
        self.deferrable_handlers = deferrable_handlers
        self.prewalk_hooks = prewalk_hooks
        self.token_handler = token_handler

    def setup(self):
        layout_handlers = {}
        deferrable_handlers = {}
        prewalk_hooks = []
        token_handler = None

        for rule in self.rules:
            r = rule()
            if r.get('token_handler'):
                if token_handler:
                    logger.warning(
                        "rule '%s' specified a new token_handler '%s', "
                        "overriding previously assigned token_handler '%s'",
                        rule.__name__, r['token_handler'].__name__,
                        token_handler.__name__,
                    )
                else:
                    logger.debug(
                        "rule '%s' specified a token_handler '%s'",
                        rule.__name__, r['token_handler'].__name__,
                    )
                token_handler = r['token_handler']
            layout_handlers.update(r.get('layout_handlers', {}))
            deferrable_handlers.update(r.get('deferrable_handlers', {}))
            prewalk_hooks.extend(r.get('prewalk_hooks', []))

        if self.token_handler:
            if token_handler and token_handler is not self.token_handler:
                logger.info(
                    "manually specified token_handler '%s' to the '%s' "
                    "instance will override rule derived token_handler '%s'",
                    self.token_handler.__name__, self.__class__.__name__,
                    token_handler.__name__
                )
            else:
                logger.debug(
                    "'%s' instance using manually specified token_handler "
                    "'%s'; ",
                    self.__class__.__name__, self.token_handler.__name__
                )
            token_handler = self.token_handler
        elif self.token_handler is None and token_handler is None:
            token_handler = token_handler_str_default
            logger.debug(
                "'%s' instance has no token_handler specified; "
                "default handler '%s' activated",
                self.__class__.__name__, token_handler.__name__
            )

        if self.layout_handlers:
            layout_handlers.update(self.layout_handlers)

        if self.deferrable_handlers:
            deferrable_handlers.update(self.deferrable_handlers)

        if self.prewalk_hooks:
            prewalk_hooks.extend(self.prewalk_hooks)

        return (
            token_handler, layout_handlers, deferrable_handlers, prewalk_hooks)

    def __call__(self, node):
        (token_handler, layout_handlers, deferrable_handlers,
            prewalk_hooks) = self.setup()
        dispatcher = self.dispatcher_cls(
            self.definitions,
            token_handler,
            layout_handlers,
            deferrable_handlers,
        )

        for prewalk_hook in prewalk_hooks:
            node = prewalk_hook(dispatcher, node)

        for chunk in self.walk(dispatcher, node):
            yield chunk



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
import textwrap
from functools import partial

from calmjs.parse import asttypes
from calmjs.parse import es5
from calmjs.parse.ruletypes import Declare
from calmjs.parse.ruletypes import Space
from calmjs.parse.ruletypes import RequiredSpace
from calmjs.parse.ruletypes import Text
from calmjs.parse.parsers.es5 import parse
from calmjs.parse.walkers import Walker

from calmjs.parse.handlers.core import layout_handler_space_drop
from calmjs.parse.handlers.core import default_rules
from calmjs.parse.handlers.core import minimum_rules
from calmjs.parse.handlers.indentation import indent

from calmjs.parse.unparsers.es5 import Unparser
from calmjs.parse.unparsers.es5 import definitions
from calmjs.parse.unparsers.es5 import pretty_print
from calmjs.parse.unparsers.es5 import minify_printer
from calmjs.parse.unparsers.es5 import minify_print

from calmjs.parse.testing.util import build_equality_testcase


def quad(items):
    return [i[:4] for i in items]


class BaseUnparserTestCase(unittest.TestCase):
    # Many of these tests are here are for showing individual fixes that
    # were done to other classes in order to properly support the source
    # map feature.

    def test_empty_program(self):
        unparser = Unparser()
        ast = parse('')
        self.assertEqual(quad(unparser(ast)), [
        ])

    def test_basic_integer(self):
        unparser = Unparser()
        ast = parse('0;')
        self.assertEqual(quad(unparser(ast)), [
            ('0', 1, 1, None),
            (';', 1, 2, None),
            ('\n', 0, 0, None),
        ])

    def test_basic_var_space_standard(self):
        unparser = Unparser()
        ast = parse('var x = 0;')
        self.assertEqual(quad(unparser(ast)), [
            ('var', 1, 1, None), (' ', 0, 0, None), ('x', 1, 5, None),
            (' ', 0, 0, None), ('=', 1, 7, None), (' ', 0, 0, None),
            ('0', 1, 9, None), (';', 1, 10, None),
            ('\n', 0, 0, None),
        ])

    def test_basic_var_decl(self):
        declared_vars = []

        def declare(dispatcher, node):
            declared_vars.append(node.value)

        unparser = Unparser(deferrable_handlers={
            Declare: declare,
        })
        ast = parse('var x = 0;')
        # just run through the ast
        quad(unparser(ast))
        self.assertEqual(['x'], declared_vars)

    def test_basic_var_space_drop(self):
        unparser = Unparser(layout_handlers={
            Space: layout_handler_space_drop,
            RequiredSpace: layout_handler_space_drop,
        })
        ast = parse('var x = 0;\nvar y = 0;')
        self.assertEqual(quad(unparser(ast)), [
            ('var', 1, 1, None), (' ', None, None, None), ('x', 1, 5, None),
            (' ', None, None, None), ('=', 1, 7, None),
            (' ', None, None, None), ('0', 1, 9, None), (';', 1, 10, None),
            ('\n', 0, 0, None),
            ('var', 2, 1, None), (' ', None, None, None), ('y', 2, 5, None),
            (' ', None, None, None), ('=', 2, 7, None),
            (' ', None, None, None), ('0', 2, 9, None), (';', 2, 10, None),
            ('\n', 0, 0, None),
        ])

    def test_force_handler_drop(self):
        # if there are no rules provided, there will also be no layout
        # handlers - not very useful as note that there is now no
        # separation between `var` and `x`.
        unparser = Unparser(rules=())
        ast = parse('var x = 0;')
        self.assertEqual(quad(unparser(ast)), [
            ('var', 1, 1, None), ('x', 1, 5, None), ('=', 1, 7, None),
            ('0', 1, 9, None),
        ])

    def test_simple_identifier(self):
        unparser = Unparser()
        ast = parse('this;')
        self.assertEqual(quad(unparser(ast)), [
            ('this', 1, 1, None), (';', 1, 5, None), ('\n', 0, 0, None),
        ])

    def test_simple_identifier_unmapped(self):
        # if the definition contains unmapped entries
        new_definitions = {}
        new_definitions.update(definitions)
        new_definitions['This'] = (Text(value='this', pos=None),)
        unparser = Unparser(definitions=new_definitions)
        ast = parse('this;')
        self.assertEqual(quad(unparser(ast)), [
            ('this', None, None, None), (';', 1, 5, None), ('\n', 0, 0, None),
        ])

    def test_empty_object(self):
        unparser = Unparser()
        ast = parse('thing = {};')
        self.assertEqual(quad(unparser(ast)), [
            ('thing', 1, 1, None), (' ', 0, 0, None), ('=', 1, 7, None),
            (' ', 0, 0, None), ('{', 1, 9, None), ('}', 1, 10, None),
            (';', 1, 11, None), ('\n', 0, 0, None),
        ])

    def test_simple_function_declare(self):
        unparser = Unparser()
        ast = parse('function foo(){}')
        self.assertEqual(quad(unparser(ast)), [
            ('function', 1, 1, None), (' ', 0, 0, None), ('foo', 1, 10, None),
            ('(', 1, 13, None), (')', 1, 14, None), (' ', 0, 0, None),
            ('{', 1, 15, None), ('\n', 0, 0, None), ('}', 1, 16, None),
            ('\n', 0, 0, None),
        ])

    def test_simple_function_invoke(self):
        unparser = Unparser()
        ast = parse('foo();')
        self.assertEqual(quad(unparser(ast)), [
            ('foo', 1, 1, None), ('(', 1, 4, None), (')', 1, 5, None),
            (';', 1, 6, None), ('\n', 0, 0, None),
        ])

    def test_new_new(self):
        unparser = Unparser()
        ast = parse('new new T();')
        self.assertEqual(quad(unparser(ast)), [
            ('new', 1, 1, None), (' ', 0, 0, None),
            ('new', 1, 5, None), (' ', 0, 0, None),
            ('T', 1, 9, None), ('(', 1, 10, None), (')', 1, 11, None),
            (';', 1, 12, None), ('\n', 0, 0, None),
        ])

    def test_getter(self):
        unparser = Unparser()
        ast = parse('x = {get p() {}};')
        self.assertEqual(quad(unparser(ast)), [
            ('x', 1, 1, None), (' ', 0, 0, None), ('=', 1, 3, None),
            (' ', 0, 0, None), ('{', 1, 5, None), ('\n', 0, 0, None),
            ('get', 1, 6, None), (' ', 0, 0, None), ('p', 1, 10, None),
            ('(', 1, 11, None), (')', 1, 12, None),
            (' ', 0, 0, None), ('{', 1, 14, None), ('\n', 0, 0, None),
            ('}', 1, 15, None), ('\n', 0, 0, None),
            ('}', 1, 16, None), (';', 1, 17, None), ('\n', 0, 0, None),
        ])

    def test_setter(self):
        unparser = Unparser()
        ast = parse('x = {set p(a) {}};')
        self.assertEqual(quad(unparser(ast)), [
            ('x', 1, 1, None), (' ', 0, 0, None), ('=', 1, 3, None),
            (' ', 0, 0, None), ('{', 1, 5, None), ('\n', 0, 0, None),
            ('set', 1, 6, None), (' ', 0, 0, None), ('p', 1, 10, None),
            ('(', 1, 11, None), ('a', 1, 12, None), (')', 1, 13, None),
            (' ', 0, 0, None), ('{', 1, 15, None), ('\n', 0, 0, None),
            ('}', 1, 16, None), ('\n', 0, 0, None),
            ('}', 1, 17, None), (';', 1, 18, None), ('\n', 0, 0, None),
        ])

    def test_switch_case_default_case(self):
        unparser = Unparser()
        ast = parse('switch (v) { case true: break; default: case false: }')
        self.assertEqual(quad(unparser(ast)), [
            ('switch', 1, 1, None), (' ', 0, 0, None), ('(', 1, 8, None),
            ('v', 1, 9, None), (')', 1, 10, None), (' ', 0, 0, None),
            ('{', 1, 12, None),
            ('\n', 0, 0, None),
            ('case', 1, 14, None), (' ', 0, 0, None), ('true', 1, 19, None),
            (':', 1, 23, None),
            ('\n', 0, 0, None),
            ('break', 1, 25, None), (';', 1, 30, None),
            ('\n', 0, 0, None),
            ('default', 1, 32, None), (':', 1, 39, None),
            ('\n', 0, 0, None),
            ('case', 1, 41, None), (' ', 0, 0, None), ('false', 1, 46, None),
            (':', 1, 51, None),
            ('\n', 0, 0, None),
            ('}', 1, 53, None),
            ('\n', 0, 0, None),
        ])

    def test_elision_0(self):
        # basically empty list
        unparser = Unparser()
        ast = parse('[];')
        self.assertEqual(quad(unparser(ast)), [
            ('[', 1, 1, None), (']', 1, 2, None),
            (';', 1, 3, None), ('\n', 0, 0, None),
        ])

    def test_elision_1(self):
        unparser = Unparser()
        ast = parse('[,];')
        self.assertEqual(quad(unparser(ast)), [
            ('[', 1, 1, None), (',', 1, 2, None), (']', 1, 3, None),
            (';', 1, 4, None), ('\n', 0, 0, None),
        ])

    def test_elision_2(self):
        unparser = Unparser()
        ast = parse('[,,];')
        self.assertEqual(quad(unparser(ast)), [
            ('[', 1, 1, None), (',,', 1, 2, None), (']', 1, 4, None),
            (';', 1, 5, None), ('\n', 0, 0, None),
        ])

    def test_elision_4(self):
        unparser = Unparser()
        ast = parse('[,,,,];')
        self.assertEqual(quad(unparser(ast)), [
            ('[', 1, 1, None), (',,,,', 1, 2, None), (']', 1, 6, None),
            (';', 1, 7, None), ('\n', 0, 0, None),
        ])

    def test_elision_v3(self):
        unparser = Unparser()
        ast = parse('[1,,,,];')
        self.assertEqual(quad(unparser(ast)), [
            ('[', 1, 1, None), ('1', 1, 2, None), (',', 0, 0, None),
            (',,,', 1, 4, None), (']', 1, 7, None),
            (';', 1, 8, None), ('\n', 0, 0, None),
        ])

    def test_elision_vv3(self):
        unparser = Unparser()
        ast = parse('[1, 2,,,,];')
        self.assertEqual(quad(unparser(ast)), [
            ('[', 1, 1, None),
            ('1', 1, 2, None), (',', 0, 0, None), (' ', 0, 0, None),
            ('2', 1, 5, None), (',', 0, 0, None),  # ditto for this
            (',,,', 1, 7, None), (']', 1, 10, None),
            (';', 1, 11, None), ('\n', 0, 0, None),
        ])

    def test_elision_v3v(self):
        unparser = Unparser()
        ast = parse('[1,,,, 1];')
        self.assertEqual(quad(unparser(ast)), [
            ('[', 1, 1, None), ('1', 1, 2, None), (',', 0, 0, None),
            (',,,', 1, 4, None),
            (' ', 0, 0, None),
            ('1', 1, 8, None), (']', 1, 9, None),
            (';', 1, 10, None), ('\n', 0, 0, None),
        ])

    def test_elision_1v1(self):
        unparser = Unparser()
        ast = parse('[, 1,,];')
        self.assertEqual(list(i[:4] for i in unparser(ast)), [
            ('[', 1, 1, None),
            (',', 1, 2, None), (' ', 0, 0, None),
            ('1', 1, 4, None), (',', 0, 0, None),
            (',', 1, 6, None), (']', 1, 7, None),
            (';', 1, 8, None), ('\n', 0, 0, None),
        ])

    def test_elision_splits_spaces(self):
        unparser = Unparser()
        # note the spaces and how they are ignored
        ast = parse('[, 1, , 2 , ,,,,, 3,, ,,,, ,,,4,];')
        self.assertEqual(list(i[:4] for i in unparser(ast)), [
            ('[', 1, 1, None), (',', 1, 2, None), (' ', 0, 0, None),
            ('1', 1, 4, None), (',', 0, 0, None),
            (',', 1, 7, None), (' ', 0, 0, None),
            ('2', 1, 9, None), (',', 0, 0, None),
            (',,,,,', 1, 13, None), (' ', 0, 0, None),
            ('3', 1, 19, None), (',', 0, 0, None),
            # though rowcols of the starting comma are maintained.
            (',,,,,,,,', 1, 21, None), (' ', 0, 0, None),
            ('4', 1, 31, None),
            (']', 1, 33, None),
            (';', 1, 34, None),
            ('\n', 0, 0, None),
        ])

    def test_elision_splits_newlines(self):
        # the newlines in this case will be completely dropped, but
        # again as long as the first comma is not shifted, it will be
        # a syntactically accurate reconstruction, while the sourcemap
        # that is generated isn't completely reflective of what the
        # original is.
        unparser = Unparser()
        ast = parse(textwrap.dedent('''
        [, 1,
        , 2 , ,,
        ,,, 3,,
        ,,,,
        ,,,4,];
        ''').strip())
        self.assertEqual(list(i[:4] for i in unparser(ast)), [
            ('[', 1, 1, None), (',', 1, 2, None), (' ', 0, 0, None),
            ('1', 1, 4, None), (',', 0, 0, None),
            (',', 2, 1, None), (' ', 0, 0, None),
            ('2', 2, 3, None), (',', 0, 0, None), (',,,,,', 2, 7, None),
            (' ', 0, 0, None),
            ('3', 3, 5, None),
            (',', 0, 0, None),
            (',,,,,,,,', 3, 7, None),
            (' ', 0, 0, None),
            ('4', 5, 4, None),
            (']', 5, 6, None),
            (';', 5, 7, None),
            ('\n', 0, 0, None),
        ])

    def test_if_else_block(self):
        unparser = Unparser()
        ast = parse('if (true) {} else {}')
        self.assertEqual([tuple(t[:4]) for t in (unparser(ast))], [
            ('if', 1, 1, None),
            (' ', 0, 0, None),
            ('(', 1, 4, None),
            ('true', 1, 5, None),
            (')', 1, 9, None),
            (' ', 0, 0, None),
            ('{', 1, 11, None),
            ('\n', 0, 0, None),
            ('}', 1, 12, None),
            ('\n', 0, 0, None),
            ('else', 1, 14, None),
            (' ', 0, 0, None),
            ('{', 1, 19, None),
            ('\n', 0, 0, None),
            ('}', 1, 20, None),
            ('\n', 0, 0, None),
        ])


class OtherUsageTestCase(unittest.TestCase):
    """
    Test out other forms of usage that are not part of the standard
    chain of calls, e.g. calls that involve manual creation/modification
    of Nodes within the AST.
    """

    def test_manual_element(self):
        # a form of possible manual replacement call.
        ast = asttypes.ES5Program(children=[
            asttypes.ExprStatement(asttypes.FunctionCall(
                identifier=asttypes.Identifier('foo'),
                args=asttypes.Arguments([asttypes.Identifier('x')]),
            )),
        ])
        unparser = Unparser()
        self.assertEqual([tuple(t) for t in (unparser(ast))], [
            ('foo', None, None, None, NotImplemented),
            ('(', None, None, None, NotImplemented),
            ('x', None, None, None, NotImplemented),
            (')', None, None, None, NotImplemented),
            (';', None, None, None, None),
            ('\n', 0, 0, None, None),
        ])

    def test_remap_function_call(self):
        # a form of possible manual replacement call.
        walker = Walker()
        src = textwrap.dedent("""
        (function(foo, bar, arg1, arg2) {
            foo(arg1);
            bar(arg2);
        })(foo, bar, arg1, arg2);
        """).strip()
        ast = parse(src)
        block = walker.extract(ast, lambda n: isinstance(n, asttypes.FuncExpr))

        for stmt in block.elements:
            fc = stmt.expr
            stmt.expr = asttypes.FunctionCall(
                args=fc.args, identifier=asttypes.DotAccessor(
                    node=asttypes.Identifier(value='window'),
                    identifier=fc.identifier))

        # Now try to render.
        unparser = Unparser()
        self.assertEqual([tuple(t[:4]) for t in (unparser(ast))], [
            ('(', 1, 1, None),
            ('function', 1, 2, None),
            ('(', 1, 10, None),
            ('foo', 1, 11, None), (',', 0, 0, None), (' ', 0, 0, None),
            ('bar', 1, 16, None), (',', 0, 0, None), (' ', 0, 0, None),
            ('arg1', 1, 21, None), (',', 0, 0, None), (' ', 0, 0, None),
            ('arg2', 1, 27, None), (')', 1, 31, None),
            (' ', 0, 0, None),
            ('{', 1, 33, None),
            ('\n', 0, 0, None),
            # injected elements should have None for lineno/colno
            ('window', None, None, None), ('.', None, None, None),
            ('foo', 2, 5, None),
            ('(', 2, 8, None), ('arg1', 2, 9, None), (')', 2, 13, None),
            (';', 2, 14, None),
            ('\n', 0, 0, None),
            ('window', None, None, None), ('.', None, None, None),
            ('bar', 3, 5, None),
            ('(', 3, 8, None), ('arg2', 3, 9, None), (')', 3, 13, None),
            (';', 3, 14, None),
            ('\n', 0, 0, None),
            ('}', 4, 1, None),
            (')', 4, 2, None),
            ('(', 4, 3, None), ('foo', 4, 4, None), (',', 0, 0, None),
            (' ', 0, 0, None), ('bar', 4, 9, None), (',', 0, 0, None),
            (' ', 0, 0, None), ('arg1', 4, 14, None), (',', 0, 0, None),
            (' ', 0, 0, None), ('arg2', 4, 20, None),
            (')', 4, 24, None), (';', 4, 25, None), ('\n', 0, 0, None),
        ])

    def test_pretty_print(self):
        # Simple test of the pretty_print function
        src = textwrap.dedent("""
        (function(foo, bar, arg1, arg2) {
          foo(arg1);
          bar(arg2);
        })(foo, bar, arg1, arg2);
        """).lstrip()
        self.assertEqual(pretty_print(parse(src)), src)

    def test_pretty_print_custom_indent(self):
        # Simple test of the pretty_print function
        src = textwrap.dedent("""
        (function(foo, bar, arg1, arg2) {
            foo(arg1);
            bar(arg2);
        })(foo, bar, arg1, arg2);
        """).lstrip()
        self.assertEqual(pretty_print(parse(src), indent_str='    '), src)

    def test_minify_print_obfuscate_skip_keywords(self):
        # test that the obfuscated minifier for es5 will skip the first
        # reserved keyword symbol found, i.e. 'do'
        # Note that while these identifiers can be easily generated by
        # a simple for loop, it can also be done by iterating through
        # an enumerated instance of the obfuscation.NameGenerator class
        # and stop at the index where the target token (in this case
        # 'do') is.
        tree = parse(textwrap.dedent("""
        (function() {
          var i_0, i_1, i_2, i_3, i_4, i_5, i_6, i_7, i_8, i_9, i_10, i_11,
            i_12, i_13, i_14, i_15, i_16, i_17, i_18, i_19, i_20, i_21, i_22,
            i_23, i_24, i_25, i_26, i_27, i_28, i_29, i_30, i_31, i_32, i_33,
            i_34, i_35, i_36, i_37, i_38, i_39, i_40, i_41, i_42, i_43, i_44,
            i_45, i_46, i_47, i_48, i_49, i_50, i_51, i_52, i_53, i_54, i_55,
            i_56, i_57, i_58, i_59, i_60, i_61, i_62, i_63, i_64, i_65, i_66,
            i_67, i_68, i_69, i_70, i_71, i_72, i_73, i_74, i_75, i_76, i_77,
            i_78, i_79, i_80, i_81, i_82, i_83, i_84, i_85, i_86, i_87, i_88,
            i_89, i_90, i_91, i_92, i_93, i_94, i_95, i_96, i_97, i_98, i_99,
            i_100, i_101, i_102, i_103, i_104, i_105, i_106, i_107, i_108,
            i_109, i_110, i_111, i_112, i_113, i_114, i_115, i_116, i_117,
            i_118, i_119, i_120, i_121, i_122, i_123, i_124, i_125, i_126,
            i_127, i_128, i_129, i_130, i_131, i_132, i_133, i_134, i_135,
            i_136, i_137, i_138, i_139, i_140, i_141, i_142, i_143, i_144,
            i_145, i_146, i_147, i_148, i_149, i_150, i_151, i_152, i_153,
            i_154, i_155, i_156, i_157, i_158, i_159, i_160, i_161, i_162,
            i_163, i_164, i_165, i_166, i_167, i_168, i_169, i_170, i_171,
            i_172, i_173, i_174, i_175, i_176, i_177, i_178, i_179, i_180,
            i_181, i_182, i_183, i_184, i_185, i_186, i_187, i_188, i_189,
            i_190, i_191, i_192, i_193, i_194, i_195, i_196, i_197, i_198,
            i_199, i_200, i_201, i_202, i_203, i_204, i_205, i_206, i_207,
            i_208, i_209, i_210, i_211, i_212, i_213, i_214, i_215, i_216,
            i_217, i_218, i_219, i_220, i_221, i_222, i_223, i_224, i_225,
            i_226 = 1;
        })();
        """))
        standard = minify_print(tree, obfuscate=False)
        self.assertIn('i_10,i_11,i_12,i_13', standard)
        minified = minify_print(tree, obfuscate=True)
        # we cannot directly test the output due to the dict lookup not
        # being deterministic, however we can assure that the bracketing
        # names (dn and dp) are generated, and ensure that 'do' is not
        # present.
        self.assertIn('dn', minified)
        self.assertNotIn('do', minified)
        self.assertIn('dp', minified)


def parse_to_sourcemap_tokens_pretty(text):
    return quad(Unparser(rules=(
        default_rules,
        indent(),
    ))(parse(text, with_comments=True)))


def parse_to_sourcemap_tokens_min(text):
    return quad(Unparser(rules=(
        minimum_rules,
    ))(parse(text)))


ParsedNodeTypeSrcmapTokenPPTestCase = build_equality_testcase(
    'ParsedNodeTypeSrcmapTokenPPTestCase', parse_to_sourcemap_tokens_pretty, ((
        label,
        textwrap.dedent(argument).strip(),
        result,
    ) for label, argument, result in [(
        'simple',
        """
        0;
        """, [
            ('0', 1, 1, None),
            (';', 1, 2, None),
            ('\n', 0, 0, None),
        ],
    ), (
        'block',
        """
        {
          var a = 5;
        }
        """, [
            ('{', 1, 1, None),
            ('\n', 0, 0, None),
            ('  ', None, None, None),
            ('var', 2, 3, None), (' ', 0, 0, None), ('a', 2, 7, None),
            (' ', 0, 0, None), ('=', 2, 9, None), (' ', 0, 0, None),
            ('5', 2, 11, None), (';', 2, 12, None),
            ('\n', 0, 0, None),
            ('}', 3, 1, None),
            ('\n', 0, 0, None),
        ],
    ), (
        'variable_statement',
        """
        var a;
        var b;
        var a, b = 3;
        var a = 1, b;
        var a = 5, b = 7;
        """, [
            ('var', 1, 1, None), (' ', 0, 0, None), ('a', 1, 5, None),
            (';', 1, 6, None),
            ('\n', 0, 0, None),
            ('var', 2, 1, None), (' ', 0, 0, None), ('b', 2, 5, None),
            (';', 2, 6, None),
            ('\n', 0, 0, None),
            ('var', 3, 1, None), (' ', 0, 0, None), ('a', 3, 5, None),
            (',', 0, 0, None), (' ', 0, 0, None),
            ('b', 3, 8, None), (' ', 0, 0, None), ('=', 3, 10, None),
            (' ', 0, 0, None), ('3', 3, 12, None), (';', 3, 13, None),
            ('\n', 0, 0, None),
            ('var', 4, 1, None), (' ', 0, 0, None), ('a', 4, 5, None),
            (' ', 0, 0, None), ('=', 4, 7, None), (' ', 0, 0, None),
            ('1', 4, 9, None), (',', 0, 0, None), (' ', 0, 0, None),
            ('b', 4, 12, None), (';', 4, 13, None),
            ('\n', 0, 0, None),
            ('var', 5, 1, None), (' ', 0, 0, None), ('a', 5, 5, None),
            (' ', 0, 0, None), ('=', 5, 7, None), (' ', 0, 0, None),
            ('5', 5, 9, None), (',', 0, 0, None), (' ', 0, 0, None),
            ('b', 5, 12, None), (' ', 0, 0, None), ('=', 5, 14, None),
            (' ', 0, 0, None), ('7', 5, 16, None), (';', 5, 17, None),
            ('\n', 0, 0, None),
        ],
    ), (
        'empty_statement',
        """
        ;
        ;
        ;
        """,
        [
            (';', 1, 1, None),
            ('\n', 0, 0, None),
            (';', 2, 1, None),
            ('\n', 0, 0, None),
            (';', 3, 1, None),
            ('\n', 0, 0, None),
        ],
    ), (
        'function_call_0',
        """
        test();
        """,
        [
            ('test', 1, 1, None), ('(', 1, 5, None), (')', 1, 6, None),
            (';', 1, 7, None),
            ('\n', 0, 0, None),
        ],
    ), (
        'function_call_1',
        """
        test(1);
        """,
        [
            ('test', 1, 1, None), ('(', 1, 5, None), ('1', 1, 6, None),
            (')', 1, 7, None),
            (';', 1, 8, None),
            ('\n', 0, 0, None),
        ],
    ), (
        'function_call_2',
        """
        test(1, 2);
        """,
        [
            ('test', 1, 1, None), ('(', 1, 5, None),
            ('1', 1, 6, None), (',', 0, 0, None), (' ', 0, 0, None),
            ('2', 1, 9, None),
            (')', 1, 10, None),
            (';', 1, 11, None),
            ('\n', 0, 0, None),
        ],
    ), (
        'operator_1',
        """
        i = 1 + 1;
        """,
        [
            ('i', 1, 1, None),
            (' ', 0, 0, None),
            ('=', 1, 3, None),
            (' ', 0, 0, None),
            ('1', 1, 5, None),
            (' ', 0, 0, None),
            ('+', 1, 7, None),
            (' ', 0, 0, None),
            ('1', 1, 9, None),
            (';', 1, 10, None),
            ('\n', 0, 0, None),
        ],
    ), (
        'unary_op',
        """
        !true;
        !1;
        delete a;
        delete(a);
        ++a;
        a++;
        """,
        [
            ('!', 1, 1, None),
            ('true', 1, 2, None),
            (';', 1, 6, None),
            ('\n', 0, 0, None),
            ('!', 2, 1, None),
            ('1', 2, 2, None),
            (';', 2, 3, None),
            ('\n', 0, 0, None),
            ('delete', 3, 1, None),
            (' ', 0, 0, None),
            ('a', 3, 8, None),
            (';', 3, 9, None),
            ('\n', 0, 0, None),
            ('delete', 4, 1, None),
            ('(', 4, 7, None),
            ('a', 4, 8, None),
            (')', 4, 9, None),
            (';', 4, 10, None),
            ('\n', 0, 0, None),
            ('++', 5, 1, None),
            ('a', 5, 3, None),
            (';', 5, 4, None),
            ('\n', 0, 0, None),
            ('a', 6, 1, None),
            ('++', 6, 2, None),
            (';', 6, 4, None),
            ('\n', 0, 0, None),
        ],
    ), (
        'object',
        """
        var obj = {
            x: 1,
            y: 2,
        }
        """,
        [
            ('var', 1, 1, None),
            (' ', 0, 0, None),
            ('obj', 1, 5, None),
            (' ', 0, 0, None),
            ('=', 1, 9, None),
            (' ', 0, 0, None),
            ('{', 1, 11, None),
            ('\n', 0, 0, None),
            ('  ', None, None, None),
            ('x', 2, 5, None),
            (':', 2, 6, None),
            (' ', 0, 0, None),
            ('1', 2, 8, None),
            (',', 3, 9, None),
            ('\n', 0, 0, None),
            ('  ', None, None, None),
            ('y', 3, 5, None),
            (':', 3, 6, None),
            (' ', 0, 0, None),
            ('2', 3, 8, None),
            ('\n', 0, 0, None),
            ('}', 4, 1, None),
            (';', 0, 0, None),
            ('\n', 0, 0, None),
        ],
    ), (
        'binop_prefixop',
        """
        var a = i+ --i;
        var b = i+ ++i;
        var c = i+ -i;
        """, [
            ('var', 1, 1, None), (' ', 0, 0, None), ('a', 1, 5, None),
            (' ', 0, 0, None), ('=', 1, 7, None), (' ', 0, 0, None),
            ('i', 1, 9, None), (' ', 0, 0, None), ('+', 1, 10, None),
            (' ', 0, 0, None), ('--', 1, 12, None), ('i', 1, 14, None),
            (';', 1, 15, None),
            ('\n', 0, 0, None),
            ('var', 2, 1, None), (' ', 0, 0, None), ('b', 2, 5, None),
            (' ', 0, 0, None), ('=', 2, 7, None), (' ', 0, 0, None),
            ('i', 2, 9, None), (' ', 0, 0, None), ('+', 2, 10, None),
            (' ', 0, 0, None), ('++', 2, 12, None), ('i', 2, 14, None),
            (';', 2, 15, None),
            ('\n', 0, 0, None),
            ('var', 3, 1, None), (' ', 0, 0, None), ('c', 3, 5, None),
            (' ', 0, 0, None), ('=', 3, 7, None), (' ', 0, 0, None),
            ('i', 3, 9, None), (' ', 0, 0, None), ('+', 3, 10, None),
            (' ', 0, 0, None), ('-', 3, 12, None), ('i', 3, 13, None),
            (';', 3, 14, None),
            ('\n', 0, 0, None),
        ],
    ), (
        'assorted_comments',
        """
        // line
        /* block
           line 2
           final */
           // hrm
           this;
         /* more? */
        this;
        """, [
            ('// line', 1, 1, None),
            ('\n', 0, 0, None),
            ('/* block\n   line 2\n   final */', 2, 1, None),
            ('\n', 0, 0, None),
            ('// hrm', 5, 4, None),
            ('\n', 0, 0, None),
            ('this', 6, 4, None),
            (';', 6, 8, None),
            ('\n', 0, 0, None),
            ('/* more? */', 7, 2, None),
            ('\n', 0, 0, None),
            ('this', 8, 1, None),
            (';', 8, 5, None),
            ('\n', 0, 0, None),
        ],
    )])
)


ParsedToMinimumTestcase = build_equality_testcase(
    'ParsedToMinimumTestcase', parse_to_sourcemap_tokens_min, ((
        label,
        textwrap.dedent(argument).strip(),
        result,
    ) for label, argument, result in [(
        'simple',
        """
        0;
        """, [
            ('0', 1, 1, None),
            (';', 1, 2, None),
        ],
    ), (
        'block',
        """
        {
          var a = 5;
        }
        """, [
            ('{', 1, 1, None),
            ('var', 2, 3, None), (' ', 0, 0, None), ('a', 2, 7, None),
            ('=', 2, 9, None), ('5', 2, 11, None), (';', 2, 12, None),
            ('}', 3, 1, None),
        ],
    ), (
        'variable_statement',
        """
        var a;
        var b;
        var a, b = 3;
        var a = 1, b;
        var a = 5, b = 7;
        """, [
            ('var', 1, 1, None), (' ', 0, 0, None), ('a', 1, 5, None),
            (';', 1, 6, None),
            ('var', 2, 1, None), (' ', 0, 0, None), ('b', 2, 5, None),
            (';', 2, 6, None),
            ('var', 3, 1, None), (' ', 0, 0, None), ('a', 3, 5, None),
            (',', 0, 0, None),
            ('b', 3, 8, None), ('=', 3, 10, None), ('3', 3, 12, None),
            (';', 3, 13, None),
            ('var', 4, 1, None), (' ', 0, 0, None), ('a', 4, 5, None),
            ('=', 4, 7, None), ('1', 4, 9, None), (',', 0, 0, None),
            ('b', 4, 12, None), (';', 4, 13, None),
            ('var', 5, 1, None), (' ', 0, 0, None), ('a', 5, 5, None),
            ('=', 5, 7, None), ('5', 5, 9, None), (',', 0, 0, None),
            ('b', 5, 12, None), ('=', 5, 14, None), ('7', 5, 16, None),
            (';', 5, 17, None),
        ],
    ), (
        'empty_statement',
        """
        ;
        ;
        ;
        """,
        [
            (';', 1, 1, None),
            (';', 2, 1, None),
            (';', 3, 1, None),
        ],
    ), (
        'function_call_0',
        """
        test();
        """,
        [
            ('test', 1, 1, None), ('(', 1, 5, None), (')', 1, 6, None),
            (';', 1, 7, None),
        ],
    ), (
        'function_call_1',
        """
        test(1);
        """,
        [
            ('test', 1, 1, None), ('(', 1, 5, None), ('1', 1, 6, None),
            (')', 1, 7, None),
            (';', 1, 8, None),
        ],
    ), (
        'function_call_2',
        """
        test(1, 2);
        """,
        [
            ('test', 1, 1, None), ('(', 1, 5, None),
            ('1', 1, 6, None), (',', 0, 0, None),
            ('2', 1, 9, None),
            (')', 1, 10, None),
            (';', 1, 11, None),
        ],
    ), (
        'operator_1',
        """
        i = 1 + 1;
        """,
        [
            ('i', 1, 1, None),
            ('=', 1, 3, None),
            ('1', 1, 5, None),
            ('+', 1, 7, None),
            ('1', 1, 9, None),
            (';', 1, 10, None),
        ],
    ), (
        'unary_op',
        """
        !true;
        !1;
        delete a;
        delete(a);
        ++a;
        a++;
        """,
        [
            ('!', 1, 1, None),
            ('true', 1, 2, None),
            (';', 1, 6, None),
            ('!', 2, 1, None),
            ('1', 2, 2, None),
            (';', 2, 3, None),
            ('delete', 3, 1, None),
            (' ', 0, 0, None),
            ('a', 3, 8, None),
            (';', 3, 9, None),
            ('delete', 4, 1, None),
            ('(', 4, 7, None),
            ('a', 4, 8, None),
            (')', 4, 9, None),
            (';', 4, 10, None),
            ('++', 5, 1, None),
            ('a', 5, 3, None),
            (';', 5, 4, None),
            ('a', 6, 1, None),
            ('++', 6, 2, None),
            (';', 6, 4, None),
        ],
    ), (
        'object',
        """
        var obj = {
            x: 1,
            y: 2,
        }
        """,
        [
            ('var', 1, 1, None),
            (' ', 0, 0, None),
            ('obj', 1, 5, None),
            ('=', 1, 9, None),
            ('{', 1, 11, None),
            ('x', 2, 5, None),
            (':', 2, 6, None),
            ('1', 2, 8, None),
            (',', 3, 9, None),
            ('y', 3, 5, None),
            (':', 3, 6, None),
            ('2', 3, 8, None),
            ('}', 4, 1, None),
            (';', 0, 0, None),
        ],
    ), (
        'binop_prefixop',
        """
        var a = i+ --i;
        var b = i+ ++i;
        var c = i+ -i;
        """, [
            ('var', 1, 1, None), (' ', 0, 0, None), ('a', 1, 5, None),
            ('=', 1, 7, None),
            ('i', 1, 9, None), ('+', 1, 10, None),
            ('--', 1, 12, None), ('i', 1, 14, None),
            (';', 1, 15, None),
            ('var', 2, 1, None), (' ', 0, 0, None), ('b', 2, 5, None),
            ('=', 2, 7, None),
            ('i', 2, 9, None), ('+', 2, 10, None),
            (' ', 0, 0, None), ('++', 2, 12, None), ('i', 2, 14, None),
            (';', 2, 15, None),
            ('var', 3, 1, None), (' ', 0, 0, None), ('c', 3, 5, None),
            ('=', 3, 7, None),
            ('i', 3, 9, None), ('+', 3, 10, None),
            ('-', 3, 12, None), ('i', 3, 13, None),
            (';', 3, 14, None),
        ],
    )])
)


ES5IdentityTestCase = build_equality_testcase(
    'ES5IdentityTestCase', es5.pretty_print, ((
        label, value, value,
    ) for label, value in ((
        label,
        # using lstrip as the pretty printer produces a trailing newline
        textwrap.dedent(value).lstrip(),
    ) for label, value in [(
        'block',
        """
        {
          var a = 5;
        }
        """,
    ), (
        'variable_statement',
        """
        var a;
        var b;
        var a, b = 3;
        var a = 1, b;
        var a = 5, b = 7;
        """,
    ), (
        'empty_statement',
        """
        ;
        ;
        ;
        """,
    ), (
        'if_statement_inline',
        """
        if (true) var x = 100;
        """,
    ), (
        'if_empty',
        """
        if (true);
        """,
    ), (
        'if_else_empty',
        """
        if (true);
        else;
        """,
    ), (
        'if_statement_block',
        """
        if (true) {
          var x = 100;
          var y = 200;
        }
        """,
    ), (
        'if_else_inline',
        """
        if (true) if (true) var x = 100;
        else var y = 200;
        """,
    ), (
        'if_else_block',
        """
        if (true) {
          var x = 100;
        }
        else {
          var y = 200;
        }
        """,
    ), (
        'if_else_if_else_block_all_empty',
        """
        if (true) {
        }
        else if (null) {
        }
        else {
        }
        """,
    ), (
        'if_else_if_else_block_nested',
        """
        if (true) {
        }
        else if (null) {
          if (true) {
          }
          else if (null) {
          }
          else {
          }
        }
        else {
        }
        """,
    ), (
        'iteration_reuse',
        """
        for (i = 0; i < 10; i++) {
          x = 10 * i;
        }
        """,
    ), (
        'iteration_var',
        """
        for (var i = 0; i < 10; i++) {
          x = 10 * i;
        }
        """,
    ), (
        'iteration_multi_value',
        """
        for (i = 0, j = 10; i < j && j < 15; i++, j++) {
          x = i * j;
        }
        """,
    ), (
        'iteration_multi_var_value',
        """
        for (var i = 0, j = 10; i < j && j < 15; i++, j++) {
          x = i * j;
        }
        """,
    ), (
        'iteration_in',
        """
        for (p in obj) {
        }
        """,
    ), (
        # retain the semicolon in the initializer part of a 'for' statement
        'iteration_conditional_initializer',
        """
        for (Q || (Q = []); d < b;) {
          d = 1;
        }
        """,
    ), (
        'iteration_new_object',
        """
        for (new Foo(); d < b;) {
          d = 1;
        }
        """,
    ), (
        'iteration_ternary_initializer',
        """
        for (2 >> (foo ? 32 : 43) && 54; 21;) {
          a = c;
        }
        """,
    ), (
        'iteration_regex_initializer',
        """
        for (/^.+/g; cond(); ++z) {
          ev();
        }
        """,
    ), (
        'iteration_var_in_obj',
        """
        for (var p in obj) {
          p = 1;
        }
        """,
    ), (
        'iteration_do_while',
        """
        do {
          x += 1;
        } while (true);
        """,
    ), (
        'while_loop',
        """
        while (false) {
          x = null;
        }
        """,
    ), (
        'while_empty',
        """
        while (false);
        """,
    ), (

        'while_loop_continue',
        """
        while (true) {
          continue;
          s = 'I am not reachable';
        }
        """,
    ), (
        'while_loop_continue_label',
        """
        while (true) {
          continue label1;
          s = 'I am not reachable';
        }
        """,
    ), (
        'while_loop_break',
        """
        while (true) {
          break;
          s = 'I am not reachable';
        }
        """,
    ), (
        'while_loop_break_label',
        """
        while (true) {
          break label1;
          s = 'I am not reachable';
        }
        """,
    ), (
        'return_empty',
        """
        {
          return;
        }
        """,
    ), (
        'return_1',
        """
        {
          return 1;
        }
        """,
    ), (
        'with_statement',
        """
        with (x) {
          var y = x * 2;
        }
        """,
    ), (
        'labelled_statement',
        """
        label: while (true) {
          x *= 3;
        }
        """,

    ), (
        'switch_statement',
        """
        switch (day_of_week) {
          case 6:
          case 7:
            x = 'Weekend';
            break;
          case 1:
            x = 'Monday';
            break;
          default:
            break;
        }
        """,

    ), (
        'switch_statement_case_default_case',
        """
        switch (result) {
          case 'good':
            do_good();
          case 'pass':
            do_pass();
            break;
          default:
            log_unexpected_result();
          case 'error':
            handle_error();
        }
        """,

    ), (
        'switch_statement_empty',
        """
        switch (a) {
          default:
          case 1:
          case 2:
        }
        """,

    ), (
        'throw_statement',
        """
        throw 'exc';
        """,

    ), (
        'debugger_statement',
        """
        debugger;
        """,

    ), (
        'expression_statement',
        """
        5 + 7 - 20 * 10;
        ++x;
        --x;
        x++;
        x--;
        x = 17 /= 3;
        s = mot ? z : /x:3;x<5;y</g / i;
        """,

    ), (
        'try_catch_statement',
        """
        try {
          x = 3;
        }
        catch (exc) {
          x = exc;
        }
        """,

    ), (
        'try_finally_statement',
        """
        try {
          x = 3;
        }
        finally {
          x = null;
        }
        """,

    ), (
        'try_catch_finally_statement',
        """
        try {
          x = 5;
        }
        catch (exc) {
          x = exc;
        }
        finally {
          y = null;
        }
        """,

    ), (
        'function_with_arguments',
        """
        function foo(x, y) {
          z = 10;
          return x + y + z;
        }
        """,

    ), (
        'function_without_arguments',
        """
        function foo() {
          return 10;
        }
        """,

    ), (
        'var_function_without_arguments',
        """
        var a = function() {
          return 10;
        };
        """,
    ), (
        'var_function_with_arguments',
        """
        var a = function foo(x, y) {
          return x + y;
        };
        """,
    ), (
        # nested function declaration
        'function_nested_declaration',
        """
        function foo() {
          function bar() {
          }
        }
        """,

    ), (
        'function_immediate_execution',
        """
        var mult = function(x) {
          return x * 10;
        }();
        """,

    ), (
        'function_call',
        """
        foo();
        """,
    ), (
        'function_call_argument',
        """
        foo(x, 7);
        """,
    ), (
        'function_call_access_element',
        """
        foo()[10];
        """,
    ), (
        'function_call_access_attribute',
        """
        foo().foo;
        """,
    ), (
        'new_keyword',
        """
        var foo = new Foo();
        """,
    ), (
        # dot accessor
        'new_keyword_dot_accessor',
        """
        var bar = new Foo.Bar();
        """,

    ), (
        # bracket accessor
        'new_keyword_bracket_accessor',
        """
        var bar = new Foo.Bar()[7];
        """,

    ), (
        # object literal
        'object_literal_litearl_keys',
        """
        var obj = {
          foo: 10,
          bar: 20
        };
        """,
    ), (
        'object_literal_numeric_keys',
        """
        var obj = {
          1: 'a',
          2: 'b'
        };
        """,
    ), (
        'object_literal_string_keys',
        """
        var obj = {
          'a': 100,
          'b': 200
        };
        """,
    ), (
        'object_literal_empty',
        """
        var obj = {};
        """,
    ), (
        # array
        'array_create_access',
        """
        var a = [1, 2, 3, 4, 5];
        var res = a[3];
        """,
    ), (
        # elision
        'elision_1',
        """
        var a = [,,,];
        """,
    ), (
        'elision_2',
        """
        var a = [1,,, 4];
        """,
    ), (
        'elision_3',
        """
        var a = [1,, 3,, 5];
        """,

    ), (
        'function_definition',
        r"""
        String.prototype.foo = function(data) {
          var tmpl = this.toString();
          return tmpl.replace(/{{\s*(.*?)\s*}}/g, function(a, b) {
            var node = data;
            if (true) {
              var value = true;
            }
            else {
              var value = false;
            }
            $.each(n.split('.'), function(i, sym) {
              node = node[sym];
            });
            return node;
          });
        };
        """,
    ), (
        'dot_accessor_integer',
        """
        (0x25).toString();
        (1e3).toString();
        (25).toString();
        """,
    ), (
        'attr_accessor_integer',
        """
        0x25["toString"]();
        1e3["toString"]();
        25["toString"]();
        """,
    ), (
        'parentheses_not_removed',
        r"""
        Expr.match[type].source + (/(?![^\[]*\])(?![^\(]*\))/.source);
        """,

    ), (
        'comparison',
        """
        (options = arguments[i]) != null;
        """,

    ), (
        'regex_test',
        r"""
        return (/h\d/i).test(elem.nodeName);
        """,

    ), (
        # https://github.com/rspivak/slimit/issues/42
        'slimit_issue_42',
        """
        e.b(d) ? (a = [c.f(j[1])], e.fn.attr.call(a, d, !0)) : a = [k.f(j[1])];
        """,

    ), (
        'closure_scope',
        """
        (function() {
          x = 5;
        }());
        """,

    ), (
        'return_statement_negation',
        """
        return !(match === true || elem.getAttribute("classid") !== match);
        """,

    ), (
        'ternary_dot_accessor',
        """
        var el = (elem ? elem.ownerDocument || elem : 0).documentElement;
        """,

    ), (
        'typeof',
        """
        typeof second.length === "number";
        """,

    ), (
        'prepostfix',
        """
        i++;
        i--;
        ++i;
        --i;
        !i;
        function foo() {
          i++;
          i--;
          ++i;
          --i;
          !i;
        }
        """,

    ), (
        'shift_ops',
        """
        x << y;
        y >> x;
        function foo() {
          x << y;
          y >> x;
        }
        """,

    ), (
        'mul_ops',
        """
        x * y;
        y / x;
        x % z;
        function foo() {
          x * y;
          y / x;
          x % z;
        }
        """,

    ), (
        'various_ops',
        """
        5 + 7 - 20 * 10;
        ++x;
        --x;
        x++;
        x--;
        x = 17 /= 3;
        1 << 2;
        foo = 2 << 3;
        1 < 2;
        bar = 1 < 2;
        1 | 2;
        bar = 1 & 2;
        1 | 2 & 8;
        bar = 1 & 2 | 8;
        x ^ y;
        bar = x ^ y;
        x && y;
        bar = x && y;
        1, 2;
        """,
    ), (
        'regex_isolated',
        """
        s = mot ? z : /x:3;x<5;y</g / i;
        """,
    ), (
        # function call in FOR init
        'function_call_in_for_init',
        """
        for (o(); i < 3; i++) {
        }
        """,

    ), (
        # function call in FOR init
        'function_call_various',
        """
        a();
        a()();
        d()['test'];
        d().test;
        var i = a();
        var i = a()();
        var i = d()['test'];
        var i = d().test;
        """,

    ), (
        # https://github.com/rspivak/slimit/issues/32
        'slimit_issue_32',
        """
        Name.prototype = {
          get fullName() {
            return this.first + " " + this.last;
          },
          set fullName(name) {
            var names = name.split(" ");
            this.first = names[0];
            this.last = names[1];
          }
        };
        """,
    ), (
        'dot_accessor_on_integer',
        """
        (0).toString();
        """,
    ), (
        'var_function_named',
        """
        var x = function y() {
        };
        """,
    ), (
        'new_expr_lhs',
        """
        new T();
        new T().derp;
        """,
    ), (
        'new_new_expr',
        # var T = function(){ return function (){} }
        """
        new new T();
        var x = new new T();
        """,
    ), (
        # delete
        'delete_keyword',
        """
        var obj = {
          foo: 1
        };
        delete obj.foo;
        """,
    ), (
        'object_various',
        """
        var obj = {
          foo: 1,
          set bar(x) {
            this._bar = x + 1;
          },
          get bar() {
            return this._bar;
          }
        };
        """,
    ), (
        'void_keyword',
        """
        void 0;
        """,
    ), (
        'instanceof',
        """
        x instanceof y;
        """,
    ), (
        # membership
        'membership_in',
        """
        1 in s;
        """,
    ), (
        'for_various',
        """
        for (;;);
        for (o < (p < q);;);
        for (o == (p == q);;);
        for (o ^ (p ^ q);;);
        for (o | (p | q);;);
        for (o & (p & q);;);
        for (a ? (b ? c : d) : false;;);
        for (var x;;);
        for (var x, y, z;;);
        """,
    ), (
        'forin_various',
        """
        for (f in o < (p < q));
        for (f in o == (p == q));
        for (f in o ^ (p ^ q));
        for (f in o | (p | q));
        for (f in o & (p & q));
        for (f in a ? (b ? c : d) : false);
        for (f in x);
        for (f in x, y, z);
        """,
    ), (
        'forin_initializer_noin',
        """
        for (var x = foo() in (bah())) {
        }
        """,
    ), (
        'dot_reserved_word',
        """
        e.case;
        """,
    ), (
        'dot_reserved_word_nobf',
        """
        for (x = e.case;;);
        """,
    ), (
        'logical_or_expr_nobf',
        """
        (true || true) || (false && false);
        """,
    ), (
        'multiplicative_expr_nobf',
        """
        !0 % 1;
        """,
    ), (
        'function_expr_1',
        """
        (function(arg) {
        });
        """,
    ), (
        'octal_slimit_issue_70',
        r"""
        var x = '\071[90m%s';
        """,
    ), (
        'special_array_char_slimit_issue_82',
        r"""
        var x = ['a', '\n'];
        """,
    ), (
        'special_string_slimit_issue_82',
        r"""
        var x = '\n';
        """,
    ), (
        'for_in_without_braces',
        """
        for (index in [1, 2, 3]) index;
        """,
    ), (
        'for_loop_into_regex_slimit_issue_54',
        """
        for (index in [1, 2, 3]) /^salign$/;
        """,
    ), (
        'line_continuation_string',
        r"""
        {
          var a = "\
          ";
        }
        """,
    ), (
        'var_non_word_char_separation',
        """
        var $foo = bar;
        """,
    )]))
)


PrintWithCommentsTestCase = build_equality_testcase(
    'PrintWithCommentsTestCase', pretty_print, ((
        label, parse(value, with_comments=True), value,
    ) for label, value in ((
        label,
        # using lstrip as the pretty printer produces a trailing newline
        textwrap.dedent(value).lstrip(),
    ) for label, value in [(
        'this_keyword',
        """
        // foo
        // foo
        /* foo */
        this;
        """,
    ), (
        'before_function',
        """
        /* a comment */
        function foo() {
        }
        """,

    ), (
        'before_if',
        """
        /* a comment */
        if (foo == bar) {
        }
        """,

    ), (
        'not_quite_before_else_if',
        """
        if (somecondition) {
        }
        else /* a comment */
        if (foo == bar) {
          // also this is indented
          var baz = 1;
        }
        """,
    ), (
        'for_loop',
        """
        /* a comment */
        // more comments
        /* even more comments */
        for (i = 0; i < 10; i++) {
          var baz = 1;
        }
        """,
    ), (
        'while_loop',
        """
        /* an infinte loop */
        while (true) {
          // this is very pointless
          var baz = 1;
        }
        """,
    )]))
)


def parse_to_sourcemap_tokens_minify(text):
    return quad(minify_printer(obfuscate=True)(parse(text)))


ParsedNodeTypeSrcmapTokenMPTestCase = build_equality_testcase(
    'ParsedNodeTypeSrcmapTokenMPTestCase', parse_to_sourcemap_tokens_minify, ((
        label,
        textwrap.dedent(argument).strip(),
        result,
    ) for label, argument, result in [(
        'simple',
        """
        0;
        """, [
            ('0', 1, 1, None),
            (';', 1, 2, None),
        ],
    ), (
        'block',
        """
        {
          var a = 5;
        }
        """, [
            ('{', 1, 1, None),
            ('var', 2, 3, None), (' ', 0, 0, None), ('a', 2, 7, None),
            ('=', 2, 9, None),
            ('5', 2, 11, None), (';', 2, 12, None),
            ('}', 3, 1, None),
        ],
    ), (
        'variable_statement',
        """
        var a;
        var b;
        var a, b = 3;
        var a = 1, b;
        var a = 5, b = 7;
        """, [
            ('var', 1, 1, None), (' ', 0, 0, None), ('a', 1, 5, None),
            (';', 1, 6, None),
            ('var', 2, 1, None), (' ', 0, 0, None), ('b', 2, 5, None),
            (';', 2, 6, None),
            ('var', 3, 1, None), (' ', 0, 0, None), ('a', 3, 5, None),
            (',', 0, 0, None),
            ('b', 3, 8, None), ('=', 3, 10, None),
            ('3', 3, 12, None), (';', 3, 13, None),
            ('var', 4, 1, None), (' ', 0, 0, None), ('a', 4, 5, None),
            ('=', 4, 7, None),
            ('1', 4, 9, None), (',', 0, 0, None),
            ('b', 4, 12, None), (';', 4, 13, None),
            ('var', 5, 1, None), (' ', 0, 0, None), ('a', 5, 5, None),
            ('=', 5, 7, None),
            ('5', 5, 9, None), (',', 0, 0, None),
            ('b', 5, 12, None), ('=', 5, 14, None),
            ('7', 5, 16, None), (';', 5, 17, None),
        ],
    ), (
        'empty_statement',
        """
        ;
        ;
        ;
        """,
        [
            (';', 1, 1, None),
            (';', 2, 1, None),
            (';', 3, 1, None),
        ],
    ), (
        'function_call_0',
        """
        test();
        """,
        [
            ('test', 1, 1, None), ('(', 1, 5, None), (')', 1, 6, None),
            (';', 1, 7, None),
        ],
    ), (
        'function_call_1',
        """
        test(1);
        """,
        [
            ('test', 1, 1, None), ('(', 1, 5, None), ('1', 1, 6, None),
            (')', 1, 7, None),
            (';', 1, 8, None),
        ],
    ), (
        'function_call_2',
        """
        test(1, 2);
        """,
        [
            ('test', 1, 1, None), ('(', 1, 5, None),
            ('1', 1, 6, None), (',', 0, 0, None),
            ('2', 1, 9, None),
            (')', 1, 10, None),
            (';', 1, 11, None),
        ],
    ), (
        'operator_1',
        """
        i = 1 + 1;
        """,
        [
            ('i', 1, 1, None),
            ('=', 1, 3, None),
            ('1', 1, 5, None),
            ('+', 1, 7, None),
            ('1', 1, 9, None),
            (';', 1, 10, None),
        ],
    ), (
        'unary_op',
        """
        !true;
        !1;
        delete a;
        delete(a);
        ++a;
        a++;
        """,
        [
            ('!', 1, 1, None),
            ('true', 1, 2, None),
            (';', 1, 6, None),
            ('!', 2, 1, None),
            ('1', 2, 2, None),
            (';', 2, 3, None),
            ('delete', 3, 1, None),
            (' ', 0, 0, None),
            ('a', 3, 8, None),
            (';', 3, 9, None),
            ('delete', 4, 1, None),
            ('(', 4, 7, None),
            ('a', 4, 8, None),
            (')', 4, 9, None),
            (';', 4, 10, None),
            ('++', 5, 1, None),
            ('a', 5, 3, None),
            (';', 5, 4, None),
            ('a', 6, 1, None),
            ('++', 6, 2, None),
            (';', 6, 4, None),
        ],
    ), (
        'object',
        """
        var obj = {
            x: 1,
            y: 2,
        }
        """,
        [
            ('var', 1, 1, None),
            (' ', 0, 0, None),
            ('obj', 1, 5, None),
            ('=', 1, 9, None),
            ('{', 1, 11, None),
            ('x', 2, 5, None),
            (':', 2, 6, None),
            ('1', 2, 8, None),
            (',', 3, 9, None),
            ('y', 3, 5, None),
            (':', 3, 6, None),
            ('2', 3, 8, None),
            ('}', 4, 1, None),
            (';', 0, 0, None),
        ],
    ), (
        'binop_prefixop',
        """
        var a = i+ --i;
        var b = i+ ++i;
        var c = i+ -i;
        """, [
            ('var', 1, 1, None), (' ', 0, 0, None), ('a', 1, 5, None),
            ('=', 1, 7, None),
            ('i', 1, 9, None), ('+', 1, 10, None),
            ('--', 1, 12, None), ('i', 1, 14, None),
            (';', 1, 15, None),
            ('var', 2, 1, None), (' ', 0, 0, None), ('b', 2, 5, None),
            ('=', 2, 7, None),
            ('i', 2, 9, None), ('+', 2, 10, None),
            (' ', 0, 0, None), ('++', 2, 12, None), ('i', 2, 14, None),
            (';', 2, 15, None),
            ('var', 3, 1, None), (' ', 0, 0, None), ('c', 3, 5, None),
            ('=', 3, 7, None),
            ('i', 3, 9, None), ('+', 3, 10, None),
            ('-', 3, 12, None), ('i', 3, 13, None),
            (';', 3, 14, None),
        ],
    ), (
        'function_list',
        """
        (function main(root) {
          root.exports = [
            (function(module, exports) {
              module.exports = {};
            }),
            (function(module, exports) {
              exports.fun = 1;
            }),
          ];
        })(this);
        """, [
            ('(', 1, 1, None), ('function', 1, 2, None), (' ', 0, 0, None),
            ('main', 1, 11, None),
            ('(', 1, 15, None), ('a', 1, 16, 'root'), (')', 1, 20, None),
            ('{', 1, 22, None), ('a', 2, 3, 'root'), ('.', 2, 7, None),
            ('exports', 2, 8, None), ('=', 2, 16, None),
            ('[', 2, 18, None),
            ('(', 3, 5, None), ('function', 3, 6, None), ('(', 3, 14, None),
            ('a', 3, 15, 'module'), (',', 0, 0, None), ('b', 3, 23, 'exports'),
            (')', 3, 30, None), ('{', 3, 32, None),
            ('a', 4, 7, 'module'), ('.', 4, 13, None),
            ('exports', 4, 14, None), ('=', 4, 22, None), ('{', 4, 24, None),
            ('}', 4, 25, None), (';', 4, 26, None), ('}', 5, 5, None),
            (')', 5, 6, None),
            (',', 0, 0, None),
            ('(', 6, 5, None), ('function', 6, 6, None), ('(', 6, 14, None),
            ('b', 6, 15, 'module'), (',', 0, 0, None), ('a', 6, 23, 'exports'),
            (')', 6, 30, None), ('{', 6, 32, None),
            ('a', 7, 7, 'exports'), ('.', 7, 14, None),
            ('fun', 7, 15, None), ('=', 7, 19, None), ('1', 7, 21, None),
            (';', 7, 22, None), ('}', 8, 5, None), (')', 8, 6, None),
            (']', 9, 3, None), (';', 9, 4, None),
            ('}', 10, 1, None), (')', 10, 2, None),
            ('(', 10, 3, None), ('this', 10, 4, None), (')', 10, 8, None),
            (';', 10, 9, None),
        ],
    ), (
        'elision_function_list',
        """
        (function main(root) {
          root.exports = [
            (function(module, exports) {
              module.exports = {};
            }),,,,,,,
            (function(module, exports) {
              exports.fun = 1;
            }),,
          ];
        })(this);
        """, [
            ('(', 1, 1, None), ('function', 1, 2, None), (' ', 0, 0, None),
            ('main', 1, 11, None),
            ('(', 1, 15, None), ('a', 1, 16, 'root'), (')', 1, 20, None),
            ('{', 1, 22, None), ('a', 2, 3, 'root'), ('.', 2, 7, None),
            ('exports', 2, 8, None), ('=', 2, 16, None),
            ('[', 2, 18, None),
            ('(', 3, 5, None), ('function', 3, 6, None), ('(', 3, 14, None),
            ('a', 3, 15, 'module'), (',', 0, 0, None), ('b', 3, 23, 'exports'),
            (')', 3, 30, None), ('{', 3, 32, None),
            ('a', 4, 7, 'module'), ('.', 4, 13, None),
            ('exports', 4, 14, None), ('=', 4, 22, None), ('{', 4, 24, None),
            ('}', 4, 25, None), (';', 4, 26, None), ('}', 5, 5, None),
            (')', 5, 6, None),
            (',', 0, 0, None), (',,,,,,', 5, 8, None),
            ('(', 6, 5, None), ('function', 6, 6, None), ('(', 6, 14, None),
            ('b', 6, 15, 'module'), (',', 0, 0, None), ('a', 6, 23, 'exports'),
            (')', 6, 30, None), ('{', 6, 32, None),
            ('a', 7, 7, 'exports'), ('.', 7, 14, None),
            ('fun', 7, 15, None), ('=', 7, 19, None), ('1', 7, 21, None),
            (';', 7, 22, None), ('}', 8, 5, None), (')', 8, 6, None),
            (',', 0, 0, None), (',', 8, 8, None),
            (']', 9, 3, None), (';', 9, 4, None),
            ('}', 10, 1, None), (')', 10, 2, None),
            ('(', 10, 3, None), ('this', 10, 4, None), (')', 10, 8, None),
            (';', 10, 9, None),
        ],
    )])
)


MinifyPrintTestCase = build_equality_testcase(
    'MinifyPrintTestCase',
    partial(minify_print, obfuscate=True, shadow_funcname=True), ((
        label,
        parse(textwrap.dedent(source).strip()),
        answer,
    ) for label, source, answer in [(
        'switch_statement',
        """
        (function() {
          var result;
          switch (day_of_week) {
            case 6:
            case 7:
              result = 'Weekend';
              break;
            case 1:
              result = 'Monday';
              break;
            default:
              break;
          }
          return result
        })();
        """,
        "(function(){var a;switch(day_of_week){case 6:case 7:a='Weekend';"
        "break;case 1:a='Monday';break;default:break;}return a;})();",
    ), (
        'function_with_arguments',
        """
        function foo(x, y) {
          z = 10 + x;
          return x + y + z;
        }
        """,
        "function foo(a,b){z=10+a;return a+b+z;}",

    ), (
        'plus_plusplus_split',
        """
        var a = b+ ++c+d;
        """,
        "var a=b+ ++c+d;"
    ), (
        'minus_plusplus_join',
        """
        var a = b- ++c+d;
        """,
        "var a=b-++c+d;"
    ), (
        'object_props',
        """
        (function() {
          Name.prototype = {
            validated: function(key) {
              return token.get(key + this.last);
            },

            get fullName() {
              return this.first + ' ' + this.last;
            },

            set fullName(name) {
              var names = name.split(' ');
              this.first = names[0];
              this.last = names[1];
            }
          };
        })();
        """,
        "(function(){Name.prototype={validated:function(a){return token.get("
        "a+this.last);},get fullName(){return this.first+' '+this.last;},"
        "set fullName(b){var a=b.split(' ');this.first=a[0];this.last=a[1];}};"
        "})();"
    ), (
        'object_props_nonword',
        """
        (function() {
          Name.prototype = {
            get $dollar() {
              return this.money;
            },

            set $dollar(value) {
              this.money = value;
            }
          };
        })();
        """,
        "(function(){Name.prototype={get $dollar(){return this.money;},"
        "set $dollar(a){this.money=a;}};})();"
    ), (
        'try_catch_shadow',
        """
        (function() {
          var value = 1;
          try {
            console.log(value);
            throw Error('welp');
          }
          catch (value) {
            console.log(value);
          }
        })();
        """,
        "(function(){var a=1;try{console.log(a);throw Error('welp');}catch(a){"
        "console.log(a);}})();"
    ), (
        'for_in_a_block',
        """
        if (true) {
            for(;;);
        }
        """,
        'if(true){for(;;);}',
    ), (
        'function_dollar_sign',
        """
        (function $() {
          (function $() {
            var foo = 1;
          })()
        })();
        """,
        '(function $(){(function a(){var a=1;})();})();',
    ), (
        'line_continuation_string',
        r"""
        var a = "\
          ";
        """,
        'var a="  ";',
    ), (
        'var_non_word_char_separation',
        r"""
        var $foo = bar;
        """,
        'var $foo=bar;',
    ), (
        'return_string',
        """
        return"foo";
        """,
        'return"foo";'
    ), (
        'return_statement_negation',
        """
        return !1;
        """,
        'return!1;'
    ), (
        'return_nonword',
        """
        return $foo;
        """,
        'return $foo;'
    ), (
        'return_underscore',
        """
        return _;
        """,
        'return _;'
    ), (
        'dollar_instanceof_dollar',
        """
        foo$ instanceof $bar;
        """,
        'foo$ instanceof $bar;'
    ), (
        'while_loop_break_nonword_label',
        """
        while (1) {
          break $dollar;
        }
        """,
        'while(1){break $dollar;}',
    ), (
        'while_continue_nonword_label',
        """
        while (1) {
          continue $dollar;
        }
        """,
        'while(1){continue $dollar;}',
    ), (
        'iteration_in_nonword',
        """
        for (p in $obj) {
        }
        """,
        'for(p in $obj){}',
    ), (
        'iteration_in_nonword_pre',
        """
        for ($bling$ in $bling$bling$) {
        }
        """,
        'for($bling$ in $bling$bling$){}',
    ), (
        'iteration_in_str',
        """
        for ($bling$ in"bingbling") {
          console.log($bling$);
        }
        """,
        'for($bling$ in"bingbling"){console.log($bling$);}',
    ), (
        'case_various',
        """
        switch (foo) {
          case $dollar:
            break;
          case !1:
            break;
          case"foo":
            break;
        }
        """,
        'switch(foo){case $dollar:break;case!1:break;case"foo":break;}',
    ), (
        'throw_various',
        """
        throw $exc;
        throw!1;
        throw"exception";
        """,
        'throw $exc;throw!1;throw"exception";',
    ), (
        'new_nonword',
        """
        new $Money();
        """,
        'new $Money();',
    )])
)


def minify_drop_semi_helper(tree):
    result = minify_print(
        tree, obfuscate=True, shadow_funcname=True, drop_semi=True)
    # try to parse the result to ensure that it also is valid
    new_tree = es5(result)
    assert result == minify_print(
        new_tree, obfuscate=True, shadow_funcname=True, drop_semi=True)
    return result


MinifyDropSemiPrintTestCase = build_equality_testcase(
    'MinifyDropSemiPrintTestCase',
    minify_drop_semi_helper, ((
        label,
        parse(textwrap.dedent(source).strip()),
        answer,
    ) for label, source, answer in [(
        'switch_statement',
        """
        (function() {
          var result;
          switch (day_of_week) {
            case 6:
            case 7:
              result = 'Weekend';
              break;
            case 1:
              result = 'Monday';
              break;
            default:
              break;
          }
          return result
        })();
        """,
        "(function(){var a;switch(day_of_week){case 6:case 7:a='Weekend';"
        "break;case 1:a='Monday';break;default:break}return a})()",
    ), (
        'function_with_arguments',
        """
        function foo(x, y) {
          z = 10 + x;
          return x + y + z;
        }
        """,
        "function foo(a,b){z=10+a;return a+b+z}",

    ), (
        'plus_plusplus_split',
        """
        var a = b+ ++c+d;
        """,
        "var a=b+ ++c+d"
    ), (
        'minus_plusplus_join',
        """
        var a = b- ++c+d;
        """,
        "var a=b-++c+d"
    ), (
        'object_props',
        """
        (function() {
          Name.prototype = {
            validated: function(key) {
              return token.get(key + this.last);
            },

            get fullName() {
              return this.first + ' ' + this.last;
            },

            set fullName(name) {
              var names = name.split(' ');
              this.first = names[0];
              this.last = names[1];
            }
          };
        })();
        """,
        "(function(){Name.prototype={validated:function(a){return token.get("
        "a+this.last)},get fullName(){return this.first+' '+this.last},"
        "set fullName(b){var a=b.split(' ');this.first=a[0];this.last=a[1]}}"
        "})()"
    ), (
        'try_catch_shadow',
        """
        (function() {
          var value = 1;
          try {
            console.log(value);
            throw Error('welp');
          }
          catch (value) {
            console.log(value);
          }
        })();
        """,
        "(function(){var a=1;try{console.log(a);throw Error('welp')}catch(a){"
        "console.log(a)}})()"
    ), (
        'for_in_a_block',
        """
        if (true) {
            for(;;);
        }
        """,
        'if(true){for(;;);}',
    ), (
        'function_dollar_sign',
        """
        (function $() {
          (function $() {
            var foo = 1;
          })()
        })();
        """,
        '(function $(){(function a(){var a=1})()})()',
    ), (
        'nested_return_function',
        """
        v = function() {
            return function() {
                return function() {
                };
            };
        };
        """,
        'v=function(){return function(){return function(){}}}',
    )])
)



# -*- coding: utf-8 -*-
"""
Rule types are used for building descriptions for processing and
regenerating for a given type of AST back into a string.
"""

from __future__ import unicode_literals

from collections import namedtuple
from functools import partial

from calmjs.parse.asttypes import Elision
from calmjs.parse.asttypes import Identifier

LayoutChunk = namedtuple('LayoutChunk', [
    'rule', 'handler', 'node'])
StreamFragment = namedtuple('StreamFragment', [
    'text', 'lineno', 'colno', 'name', 'source'])


def is_empty(value):
    return value in (None, [])


class Rule(object):
    """
    The base type.
    """


class Layout(Rule):
    """
    Layout rules are simply for building markers that either return some
    extraneous fragments or set some flags on the Dispatcher object for
    further usage, so that the pretty printing will look pretty.

    The subclasses will be used as types for the handler lookup (i.e.
    they will serve as keys); for the subclasses to be used as par of
    the definition for the structure of a given asttype, it must be
    instantiated.
    """


class Structure(Layout):
    """
    These are for structural related, i.e. does not relate to production
    of output chunks.  They will be acted upon immediately.

    The resolved function for these layout tokens should have the
    following signature:

    (dispatcher, node)
    """


class Format(Layout):
    """
    These are format related layouts.  The walk function generally will
    buffer these for the final generation.

    The resolved function for these layout tokens should have the
    following signature:

    (dispatcher, node, before_text, after_text, prev_text)
    """


class Token(Rule):
    """
    Token rules are callable rules that will directly take part in
    ochestrating the rendering of the given input node, along with the
    walk function and the Dispatcher object that tracks the currently
    executing prettyprint iteration.

    Like the Layout type, Token types must also be instantiated before
    they can be used.
    """

    def __init__(self, attr=None, value=None, pos=0):
        """
        Arguments

        attr
            Should reference some attribute of the referenced Node
        value
            Some value that will be assigned to this token for use in
            the production of chunks
        pos
            The index to be passed to Token.getpos for the correct
            index for the given token map entry.  The text element will
            be used for the lookup first, and typically the correct
            position will be reported on the 0th entry, however for
            certain Node types multiple entries may be use, and so for
            that situation the pos will need to be explicitly specified,
            increment for each time that the exact value may be
            generated through the execution of the rule description for
            a given Node.
        """

        self.attr = attr
        self.value = value
        self.pos = pos

    def __call__(self, walk, dispatcher, node):
        """
        Arguments

        walk
            the walk function
        dispatcher
            a prettyprint.Dispatcher instance.
        node
            a Node instance.
        """

        raise NotImplementedError
        yield  # pragma: no cover


class Deferrable(Rule):
    """
    These are rules that can be considered as a restricted subset of
    Token types in the sense that they cannot independently have the
    ability to walk through the node (as the callable signature does not
    have provision for one), however the Dispatcher has a provision to
    acquire a specific handler for this Rule type (hence deferrable,
    i.e. the production of output can be deferred to the Dispatcher).

    An initial implementation was for the _getattr private method of the
    Attr Token type, where originally it was just a simple callable so
    that iter can be used to generate an iterator, however that is now
    formalized to be of this type.

    Another use case, as mentioned, is the ability for these rules to
    defer to the Dispatcher for the generation of output, which this
    instance can make use of to achieve that by acquiring a possible
    handler like so in the __call__ implementation:

        handler = dispatcher.deferrable(self)

    If a callable was returned, it should be invoked with the same
    arguments that were passed into that context, which should be
    (dispatcher, node).

    Note that this particular Rule type is to be defined within the
    context of a Token, as these are designed to operate within one.

    The constructors for subclasses are implementation specific.
    """

    def __call__(self, dispatcher, node):
        """
        Arguments

        dispatcher
            a walker.Dispatcher instance.
        node
            a Node instance.
        """

        raise NotImplementedError


class OpenBlock(Format):
    """
    Denotes a block being opened (in ES5, '{').
    """


class CloseBlock(Format):
    """
    Denotes a block being closed (in ES5, '}').
    """


class EndStatement(Format):
    """
    Denotes a statement ending (in ES5, ';').
    """


class Space(Format):
    """
    Represents a space character with unspecified characteristics.
    """


class OptionalSpace(Format):
    """
    Represents optional space character.
    """


class RequiredSpace(Format):
    """
    Represents a required space character.
    """


class Newline(Format):
    """
    Represents a newline character.
    """


class OptionalNewline(Format):
    """
    Represents an optional newline character.
    """


class Indent(Format):
    """
    Represents an increment to the indentation level.
    """


class Dedent(Format):
    """
    Represents an decrement to the indentation level.
    """


class PushScope(Structure):
    """
    Push in a new scope
    """


class PopScope(Structure):
    """
    Pops out a scope.
    """


class PushCatch(Structure):
    """
    Push catch context
    """


class PopCatch(Structure):
    """
    Pops catch context.
    """


class ResolveFuncName(Structure):
    """
    This special token is used for resolving the name of the function
    from inside the scope.

    This permits the handling the incorrectly implemented strict mode
    for the Safari browser in the use case of name mangling.
    """


class Attr(Token):
    """
    Yield the return value as specified by attribute, where it could be
    the name of the attribute for the Node, or deferrable.
    """

    def _getattr(self, dispatcher, node):
        if isinstance(self.attr, Deferrable):
            return self.attr(dispatcher, node)
        else:
            return getattr(node, self.attr)

    def __call__(self, walk, dispatcher, node):
        value = self._getattr(dispatcher, node)
        if is_empty(value):
            return
        for chunk in walk(dispatcher, value, token=self):
            yield chunk


class CommentsAttr(Attr):
    """
    Essentially identical to Attr, except default to handling comments
    such that the rules can be more easily filtered out if required.
    """

    def __init__(self, attr='comments', value=None, pos=0):
        super(CommentsAttr, self).__init__(attr=attr, value=value, pos=pos)


class Text(Token):
    """
    Simply leverage the dispatcher object for generating the rendering
    of the value required.
    """

    def __call__(self, walk, dispatcher, node):
        for chunk in walk(dispatcher, self.value, token=self):
            yield chunk


class JoinAttr(Attr):
    """
    Like Attr, though instead of yielding the returned value, walk each
    of the resulting nodes with the provided walk function, interspersed
    with walking the provided value.

    In effect, for a textual output, it would resemble invoking the
    `value.join` method on each of the unparsed chunks.
    """

    def __call__(self, walk, dispatcher, node):
        nodes = iter(self._getattr(dispatcher, node))

        try:
            target_node = next(nodes)
        except StopIteration:
            return

        for chunk in walk(dispatcher, target_node, token=self):
            yield chunk

        for target_node in nodes:
            # note that self.value is to be defined in the definition
            # format also - if it is None, the default lookup will be
            # done again which would effectively repeat the previous
            # walk as it will trigger a lookup!
            definition = self.value if self.value else ()
            for value_node in walk(dispatcher, node, definition=definition):
                yield value_node
            for chunk in walk(dispatcher, target_node, token=self):
                yield chunk


class ElisionToken(Attr, Text):
    """
    The special snowflake token just for Elision, simply because of how
    the ES5 specification (ECMA-262), section 11.1.4, specifies how the
    production of Elision is to be done.  The value is to be captured as
    an integer and is then later used by whatever processing required.

    While the name of this class is very specific, the meaning of the
    arguments for the constructor will remain the same.  The `attr` be
    the attribute to extract from the token, with the value being
    handled similar to the Text interpretation of value.
    """

    def __call__(self, walk, dispatcher, node):
        value = self._getattr(dispatcher, node)
        for chunk in walk(dispatcher, self.value * value, token=self):
            yield chunk


class ElisionJoinAttr(ElisionToken):
    """
    The Elision type does require a bit of special handling, given that
    particular Node type supplies the token that serves as the joiner
    for the preceding and subsequent nodes, including other Elision
    nodes which must also be handled separately.

    This Token implementation makes the assumption that the Elision
    nodes with a length of 1 result in an equivalent representation of
    separators (which typically is ',' for ES5) for the rendering of the
    nodes provided.

    Also note that the value should be a description (i.e. tuple of
    rules) that do not contain any Text tokens for generating the
    separator - that will be done through the Elision token.
    """

    # the surrogate Elision node to be treated as a separator.
    sep = Elision(1)
    sep._token_map = {}  # so its getpos returns an implied position

    def __call__(self, walk, dispatcher, node):
        nodes = iter(self._getattr(dispatcher, node))

        try:
            previous_node = next(nodes)
        except StopIteration:
            return

        for chunk in walk(dispatcher, previous_node, token=self):
            yield chunk

        for next_node in nodes:
            if not isinstance(previous_node, Elision):
                # Have the walk function walk our "separator" node, as
                # explained in the docstring for this class.
                for c in walk(dispatcher, self.sep):
                    yield c

            if not isinstance(next_node, Elision):
                for value_node in walk(dispatcher, node, self.value):
                    yield value_node
            for chunk in walk(dispatcher, next_node, token=self):
                yield chunk
            previous_node = next_node


class Optional(Token):
    """
    Optional text, depending on the node having an attribute

    attr
        the attr that is required for the statement to execute
    value
        the token definition segment to be executed
    """

    def __call__(self, walk, dispatcher, node):
        if is_empty(getattr(node, self.attr)):
            return

        # note that self.value is to be defined in the definition
        # format also.
        for chunk in walk(dispatcher, node, self.value):
            yield chunk


class Operator(Attr):
    """
    An operator symbol.
    """

    # A nice to have feature will be the ability to have this produce
    # something that marks the operator with rendering information, so
    # that the heuristics that are currently being employed for dealing
    # with things such as whitespaces be not so ad-hoc.

    def _getattr(self, dispatcher, node):
        if self.attr:
            return getattr(node, self.attr)
        else:
            return self.value


class Iter(Deferrable):
    """
    Produces an iterator for the Attr.
    """

    def __call__(self, dispatcher, node):
        return iter(node)


class Declare(Deferrable):
    """
    Record the declaration of the specific identifier.
    """

    def __init__(self, attr):
        self.attr = attr

    def _handle(self, handler, parent, child, idx=None):
        if not isinstance(child, Identifier):
            raise TypeError(
                "in %r, the resolved attribute '%s%s' is not an Identifier" %
                (parent, self.attr, '' if idx is None else '[%d]' % idx)
            )
        # invoke it with the specific method signature for Deferrable
        # handlers.
        handler(child)

    def __call__(self, dispatcher, node):
        target = getattr(node, self.attr)

        if is_empty(target):
            # can't record nothing.
            return target

        # look up the layout handler for this deferrable type
        handler = dispatcher.deferrable(self)
        if handler is not NotImplemented:
            handler = partial(handler, dispatcher)
            if isinstance(target, list):
                for idx, item in enumerate(target):
                    self._handle(handler, node, item, idx)
            else:
                self._handle(handler, node, target)

        # finally, return the actual attribute
        return target


class Resolve(Deferrable):
    """
    Resolve an identifier.  Naturally, this is used by the Identifier
    asttype to look up previously declared names.
    """

    def __call__(self, dispatcher, node):
        if not isinstance(node, Identifier):
            raise TypeError(
                "the Resolve Deferrable type only works with Identifier")

        handler = dispatcher.deferrable(self)
        if handler is not NotImplemented:
            # the handler will return the value
            return handler(dispatcher, node)
        return node.value


class Literal(Deferrable):
    """
    Provides special handling for literals such as strings.
    """

    def __call__(self, dispatcher, node):
        handler = dispatcher.deferrable(self)
        if handler is not NotImplemented:
            # the handler will return the value
            return handler(dispatcher, node)
        return node.value


class Comment(Deferrable):
    """
    Provide special handling for comments of either types.
    """

    # yes this is similar enough to Literal, but for now keep the two
    # implementation separate until merging can be proven to be harmless

    def __call__(self, dispatcher, node):
        handler = dispatcher.deferrable(self)
        if handler is not NotImplemented:
            # the handler will return the value
            return handler(dispatcher, node)
        # do not return a default for now?


class LineComment(Comment):
    """
    Provide special handling for line comments
    """


class BlockComment(Comment):
    """
    Provide special handling for line comments
    """


children_newline = JoinAttr(Iter(), value=(Newline,))
children_comma = JoinAttr(Iter(), value=(Text(value=','), Space,))



# -*- coding: utf-8 -*-
import unittest

from calmjs.parse.ruletypes import Token
from calmjs.parse.ruletypes import Layout
from calmjs.parse.ruletypes import Deferrable


class PPTypesTestCase(unittest.TestCase):
    """
    Tests for selected pretty print types.
    """

    def test_layout(self):
        Layout()

    def test_token(self):
        token = Token()
        self.assertTrue(callable(token))
        rule = token(None, None, None)
        with self.assertRaises(NotImplementedError):
            next(rule)

    def test_deferrable(self):
        deferrable = Deferrable()
        self.assertTrue(callable(deferrable))
        with self.assertRaises(NotImplementedError):
            deferrable(None, None)



# -*- coding: utf-8 -*-
"""
Various helpers for plugging into the pprint walker.

Combining the following layout handlers with the ruletypes definition
and pprint framework can be done in myriad of ways such that new output
formats can be constructed very trivially by simply changing how or what
of the following layouts to use to plug into the unparser walker setup.
"""

from __future__ import unicode_literals

import re

from calmjs.parse.asttypes import (
    Identifier,
    If,
    For,
    ForIn,
    While,
)
from calmjs.parse.ruletypes import (
    StreamFragment,

    OpenBlock,
    CloseBlock,
    EndStatement,
    Space,
    OptionalSpace,
    RequiredSpace,
    Newline,
    OptionalNewline,
    Indent,
    Dedent,
    LineComment as RuleTypeLineComment,
    BlockComment as RuleTypeBlockComment,
)
from calmjs.parse.lexers.es5 import PATT_LINE_CONTINUATION

required_space = re.compile(r'^(?:\w\w|\+\+|\-\-|\w\$|\$\w)$')

# the various assignments symbols; for dealing with pretty spacing
assignment_tokens = {
    '*=', '/=', '%=', '+=', '-=', '<<=', '>>=', '>>>=', '&=', '^=', '|=', '='}
# other symbols
optional_rhs_space_tokens = {';', ')', None}
space_imply = StreamFragment(' ', 0, 0, None, None)
space_drop = StreamFragment(' ', None, None, None, None)


def rule_handler_noop(*a, **kw):
    # a no op for layouts
    return
    yield  # pragma: no cover


def token_handler_str_default(
        token, dispatcher, node, subnode, sourcepath_stack=(None,)):
    """
    Standard token handler that will return the value, ignoring any
    tokens or strings that have been remapped.
    """

    if isinstance(token.pos, int):
        _, lineno, colno = node.getpos(subnode, token.pos)
    else:
        lineno, colno = None, None
    yield StreamFragment(subnode, lineno, colno, None, sourcepath_stack[-1])


def token_handler_unobfuscate(
        token, dispatcher, node, subnode, sourcepath_stack=(None,)):
    """
    A token handler that will resolve and return the original identifier
    value.
    """

    original = (
        node.value
        if isinstance(node, Identifier) and node.value != subnode else
        None
    )

    if isinstance(token.pos, int):
        _, lineno, colno = node.getpos(original or subnode, token.pos)
    else:
        lineno, colno = None, None

    yield StreamFragment(
        subnode, lineno, colno, original, sourcepath_stack[-1])


def layout_handler_semicolon(dispatcher, node, before, after, prev):
    # required layout handler for the EndStatement Format rule.
    _, lineno, colno = node.getpos(';', 0)
    yield StreamFragment(';', lineno, colno, None, None)


def layout_handler_semicolon_optional(dispatcher, node, before, after, prev):
    # only yield if there is something after
    if after:
        _, lineno, colno = node.getpos(';', 0)
        yield StreamFragment(';', lineno, colno, None, None)


def layout_handler_openbrace(dispatcher, node, before, after, prev):
    # required layout handler for the OpenBlock Format rule.
    _, lineno, colno = node.getpos('{', 0)
    yield StreamFragment('{', lineno, colno, None, None)


def layout_handler_closebrace(dispatcher, node, before, after, prev):
    # required layout handler for the CloseBlock Format rule.
    _, lineno, colno = node.getpos('}', 0)
    yield StreamFragment('}', lineno, colno, None, None)


def layout_handler_space_imply(dispatcher, node, before, after, prev):
    # default layout handler where the space will be rendered, with the
    # line/column set to 0 for sourcemap to generate the implicit value.
    yield space_imply


def layout_handler_space_drop(dispatcher, node, before, after, prev):
    # default layout handler where the space will be rendered, with the
    # line/column set to None for sourcemap to terminate the position.
    yield space_drop


def layout_handler_newline_simple(dispatcher, node, before, after, prev):
    # simply render the newline with an implicit sourcemap line/col
    yield StreamFragment(dispatcher.newline_str, 0, 0, None, None)


def layout_handler_newline_optional_pretty(
        dispatcher, node, before, after, prev):
    # simply render the newline with an implicit sourcemap line/col, if
    # not already preceded or followed by a newline
    idx = len(dispatcher.newline_str)

    def fc(s):
        return '' if s is None else s[:idx]

    def lc(s):
        return '' if s is None else s[-idx:]

    # include standard ones plus whatever else that was provided, i.e.
    # the typical <CR><LF>
    newline_strs = {'\r', '\n', dispatcher.newline_str}

    if lc(before) in '\r\n':
        # not needed since this is the beginning
        return
    # if no new lines in any of the checked characters
    if not newline_strs & {lc(before), fc(after), lc(prev)}:
        yield StreamFragment(dispatcher.newline_str, 0, 0, None, None)


def layout_handler_space_optional_pretty(
        dispatcher, node, before, after, prev):
    if isinstance(node, (If, For, ForIn, While)):
        if after not in optional_rhs_space_tokens:
            yield space_imply
            return

    if before is None or after is None:
        # nothing.
        return
    s = before[-1:] + after[:1]

    if required_space.match(s) or after in assignment_tokens:
        yield space_imply
        return


def layout_handler_space_minimum(dispatcher, node, before, after, prev):
    if before is None or after is None:
        # nothing.
        return
    s = before[-1:] + after[:1]
    if required_space.match(s):
        yield space_imply


def deferrable_handler_literal_continuation(dispatcher, node):
    # assume the es5 method will continue to work.
    return PATT_LINE_CONTINUATION.sub('', node.value)


def deferrable_handler_comment(dispatcher, node):
    # simply return the value
    return node.value


def default_rules():
    return {'layout_handlers': {
        OpenBlock: layout_handler_openbrace,
        CloseBlock: layout_handler_closebrace,
        EndStatement: layout_handler_semicolon,
        Space: layout_handler_space_imply,
        OptionalSpace: layout_handler_space_optional_pretty,
        RequiredSpace: layout_handler_space_imply,
        Newline: layout_handler_newline_simple,
        OptionalNewline: layout_handler_newline_optional_pretty,
        # define these as noop so they can be normalized
        Indent: rule_handler_noop,
        Dedent: rule_handler_noop,
        # if an indent is immediately followed by dedent without actual
        # content, simply do nothing.
        (Indent, Newline, Dedent): rule_handler_noop,
        (OptionalSpace, EndStatement): layout_handler_semicolon,
    }, 'deferrable_handlers': {
        RuleTypeLineComment: deferrable_handler_comment,
        RuleTypeBlockComment: deferrable_handler_comment,
    }}


def minimum_rules():
    return {'layout_handlers': {
        OpenBlock: layout_handler_openbrace,
        CloseBlock: layout_handler_closebrace,
        EndStatement: layout_handler_semicolon,
        Space: layout_handler_space_minimum,
        OptionalSpace: layout_handler_space_minimum,
        RequiredSpace: layout_handler_space_imply,
        # drop the space before '{'
        (Space, OpenBlock): layout_handler_openbrace,
        # remove space before ';' for empty for statement
        (Space, EndStatement): layout_handler_semicolon,
        (OptionalSpace, EndStatement): layout_handler_semicolon,
    }}



###############################################################################
#
# Copyright (c) 2011 Ruslan Spivak
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from __future__ import unicode_literals

__author__ = 'Ruslan Spivak <ruslan.spivak@gmail.com>'

import re
import ply.lex

from calmjs.parse.lexers.tokens import AutoLexToken
from calmjs.parse.utils import repr_compat
from calmjs.parse.exceptions import (
    ECMASyntaxError,
    ECMARegexSyntaxError,
)
from calmjs.parse.unicode_chars import (
    LETTER,
    DIGIT,
    COMBINING_MARK,
    CONNECTOR_PUNCTUATION,
)
from calmjs.parse.utils import format_lex_token

# See "Regular Expression Literals" at
# http://www.mozilla.org/js/language/js20-2002-04/rationale/syntax.html
TOKENS_THAT_IMPLY_DIVISON = frozenset([
    'ID',
    'NUMBER',
    'STRING',
    'REGEX',
    'TRUE',
    'FALSE',
    'NULL',
    'THIS',
    'PLUSPLUS',
    'MINUSMINUS',
    'RPAREN',
    'RBRACE',
    'RBRACKET',
])

IMPLIED_BLOCK_IDENTIFIER = frozenset([
    'FOR',
    'WHILE',
    'IF',
])


# think of a better name for this, but this is mostly to address section
# 7 of the spec.
DIVISION_SYNTAX_MARKERS = frozenset([
    'LINE_TERMINATOR', 'LINE_COMMENT', 'BLOCK_COMMENT'
])

COMMENTS = frozenset([
    'LINE_COMMENT', 'BLOCK_COMMENT'
])

PATT_LINE_TERMINATOR_SEQUENCE = re.compile(
    r'(\n|\r(?!\n)|\u2028|\u2029|\r\n)', flags=re.S)
PATT_LINE_CONTINUATION = re.compile(
    r'\\(\n|\r(?!\n)|\u2028|\u2029|\r\n)', flags=re.S)


PATT_BROKEN_STRING = re.compile(r"""
(?:
    # broken double quoted string
    (?:"                               # opening double quote
        (?: [^"\\\n\r\u2028\u2029]     # not ", \, line terminators; allow
            | \\(\n|\r(?!\n)|\u2028|\u2029|\r\n)  # line continuation
            | \\[a-tvwyzA-TVWYZ!-\/:-@\[-`{-~] # escaped chars
            | \\x[0-9a-fA-F]{2}        # hex_escape_sequence
            | \\u[0-9a-fA-F]{4}        # unicode_escape_sequence
            | \\(?:[1-7][0-7]{0,2}|[0-7]{2,3})  # octal_escape_sequence
            | \\0                      # <NUL> (15.10.2.11)
        )*                             # and capture them greedily
    )                                  # omit closing quote
    |
    # broken single quoted string
    (?:'                               # opening single quote
        (?: [^'\\\n\r\u2028\u2029]     # not ', \, line terminators; allow
            | \\(\n|\r(?!\n)|\u2028|\u2029|\r\n)  # line continuation
            | \\[a-tvwyzA-TVWYZ!-\/:-@\[-`{-~] # escaped chars
            | \\x[0-9a-fA-F]{2}        # hex_escape_sequence
            | \\u[0-9a-fA-F]{4}        # unicode_escape_sequence
            | \\(?:[1-7][0-7]{0,2}|[0-7]{2,3}) # octal_escape_sequence
            | \\0                      # <NUL> (15.10.2.11)
        )*                             # and capture them greedily
    )                                  # omit closing quote
)
""", flags=re.VERBOSE)


def broken_string_token_handler(lexer, token):
    match = PATT_BROKEN_STRING.match(token.value)
    if not match:
        return

    # update the error token value to only include what was matched here
    # as this will be the actual token that "failed"
    token.value = match.group()
    # calculate colno for current token colno before...
    colno = lexer._get_colno(token)
    # updating the newline indexes for the error reporting for raw
    # lexpos
    lexer._update_newline_idx(token)
    # probe for the next values (which no valid rules will match)
    position = lexer.lexer.lexpos + len(token.value)
    failure = lexer.lexer.lexdata[position:position + 2]
    if failure and failure[0] == '\\':
        type_ = {'x': 'hexadecimal', 'u': 'unicode'}[failure[1]]
        seq = re.match(
            r'\\[xu][0-9-a-f-A-F]*', lexer.lexer.lexdata[position:]
        ).group()
        raise ECMASyntaxError(
            "Invalid %s escape sequence '%s' at %s:%s" % (
                type_, seq, lexer.lineno,
                lexer._get_colno_lexpos(position)
            )
        )
    tl = 16  # truncate length
    raise ECMASyntaxError(
        'Unterminated string literal %s at %s:%s' % (
            repr_compat(
                token.value[:tl].strip() + (token.value[tl:] and '...')),
            token.lineno, colno)
    )


class Lexer(object):
    """A JavaScript lexer.

    >>> from calmjs.parse.lexers.es5 import Lexer
    >>> lexer = Lexer()

    Lexer supports iteration:

    >>> lexer.input('a = 1;')
    >>> for token in lexer:
    ...     print(str(token).replace("u'", "'"))
    ...
    LexToken(ID,'a',1,0)
    LexToken(EQ,'=',1,2)
    LexToken(NUMBER,'1',1,4)
    LexToken(SEMI,';',1,5)

    Or call one token at a time with 'token' method:

    >>> lexer.input('a = 1;')
    >>> while True:
    ...     token = lexer.token()
    ...     if not token:
    ...         break
    ...     print(str(token).replace("u'", "'"))
    ...
    LexToken(ID,'a',1,0)
    LexToken(EQ,'=',1,2)
    LexToken(NUMBER,'1',1,4)
    LexToken(SEMI,';',1,5)

    >>> lexer.input('a = 1;')
    >>> token = lexer.token()
    >>> str(token.type), str(token.value), token.lineno, token.lexpos
    ('ID', 'a', 1, 0)

    For more information see:
    http://www.ecma-international.org/publications/files/ECMA-ST/ECMA-262.pdf
    """
    def __init__(self, with_comments=False, yield_comments=False):
        self.lexer = None
        self.prev_token = None
        # valid_prev_token is for syntax error hint, and also for
        # tracking real tokens
        self.valid_prev_token = None
        self.cur_token = None
        self.cur_token_real = None
        self.next_tokens = []
        self.token_stack = [[None, []]]
        self.newline_idx = [0]
        self.error_token_handlers = [
            broken_string_token_handler,
        ]
        self.with_comments = with_comments
        self.yield_comments = yield_comments
        self.hidden_tokens = []
        self.build()

        if not with_comments:
            # just reassign the method.
            self.token = self._token

    @property
    def lineno(self):
        return self.lexer.lineno if self.lexer else 0

    @property
    def lexpos(self):
        return self.lexer.lexpos if self.lexer else 0

    @property
    def last_newline_lexpos(self):
        return self.newline_idx[-1]

    def build(self, **kwargs):
        """Build the lexer."""
        self.lexer = ply.lex.lex(object=self, **kwargs)

    def input(self, text):
        self.lexer.input(text)

    def _update_newline_idx(self, token):
        lexpos = token.lexpos
        fragments = PATT_LINE_TERMINATOR_SEQUENCE.split(token.value)
        for fragment, newline in zip(*[iter(fragments)] * 2):
            lexpos += len(fragment + newline)
            self.lexer.lineno += 1
            self.newline_idx.append(lexpos)

    def get_lexer_token(self):
        token = self.lexer.token()
        if token:
            token.colno = self._get_colno(token)
            self._update_newline_idx(token)
        return token

    def backtracked_token(self, pos=1):
        self.lexer.skip(- pos)
        # clearly the buffer here needs wiping too
        self.next_tokens = []
        # do the dance to ensure the valid previous tokens are tracked.
        valid_prev_token = self.valid_prev_token
        token = self.token()
        self.valid_prev_token = valid_prev_token
        return token

    def token(self):
        token = self._token()
        if token and self.hidden_tokens:
            token.hidden_tokens = self.hidden_tokens
            self.hidden_tokens = []
        return token

    def _token(self):
        # auto-semi tokens that got added
        if self.next_tokens:
            return self.next_tokens.pop()

        lexer = self.lexer
        while True:
            pos = lexer.lexpos
            try:
                char = lexer.lexdata[pos]
                while char in ' \t':
                    pos += 1
                    char = lexer.lexdata[pos]
                next_char = lexer.lexdata[pos + 1]
            except IndexError:
                tok = self._get_update_token()
                if tok is not None and tok.type == 'LINE_TERMINATOR':
                    # should this also be implemented?
                    # if not self.drop_lineterm:
                    #     self.hidden_tokens.append(tok)
                    continue
                else:
                    return tok

            if char != '/' or (char == '/' and next_char in ('/', '*')):
                tok = self._get_update_token()
                if tok.type in DIVISION_SYNTAX_MARKERS:
                    if tok.type in COMMENTS:
                        if self.yield_comments:
                            return tok
                        elif self.with_comments:
                            self.hidden_tokens.append(tok)
                    continue
                else:
                    return tok

            # current character is '/' which is either division or regex
            # First check that if the previous token is a potential
            # division syntax marker that matches section 7.  If so, use
            # the previous token.
            check_token = (
                self.cur_token_real
                if self.cur_token_real is None or
                self.cur_token_real.type not in DIVISION_SYNTAX_MARKERS else
                self.prev_token
            )
            is_division_allowed = (
                check_token is not None and
                check_token.type in TOKENS_THAT_IMPLY_DIVISON
            ) and (
                self.token_stack[-1][0] is None or (
                    # if the token on the stack is the same, the
                    # following was done already, so skip it
                    self.token_stack[-1][0] is self.prev_token or
                    self.token_stack[-1][0].type in TOKENS_THAT_IMPLY_DIVISON
                )
            )
            if is_division_allowed:
                return self._get_update_token()
            else:
                self._set_tokens(self._read_regex())
                return self.cur_token

    def auto_semi(self, token):
        if token is None or (token.type not in ('SEMI', 'AUTOSEMI') and (
                token.type == 'RBRACE' or self._is_prev_token_lt())):
            if token:
                self.next_tokens.append(token)
            return self._create_semi_token(token)

    def _set_tokens(self, new_token):
        self.token_stack[-1][0] = self.prev_token = self.cur_token
        if (self.cur_token and
                self.cur_token.type not in DIVISION_SYNTAX_MARKERS):
            self.valid_prev_token = self.cur_token
        self.cur_token = new_token
        if (self.cur_token and
                self.cur_token.type not in DIVISION_SYNTAX_MARKERS):
            self.cur_token_real = self.cur_token

    def _is_prev_token_lt(self):
        return self.prev_token and self.prev_token.type == 'LINE_TERMINATOR'

    def _read_regex(self):
        self.lexer.begin('regex')
        token = self.get_lexer_token()
        self.lexer.begin('INITIAL')
        return token

    def _get_update_token(self):
        self._set_tokens(self.get_lexer_token())

        if self.cur_token is not None:

            if self.cur_token.type in ('LPAREN',):
                # if we encounter a FOR, IF, WHILE, then whatever in
                # the parentheses are marked.  Otherwise just push
                # into the inner marker list.
                if (self.prev_token and
                        self.prev_token.type in IMPLIED_BLOCK_IDENTIFIER):
                    self.token_stack.append([self.cur_token, []])
                else:
                    self.token_stack[-1][1].append(self.cur_token)

            if self.cur_token.type in ('RPAREN',):
                # likewise, pop the inner marker first.
                if self.token_stack[-1][1]:
                    self.token_stack[-1][1].pop()
                else:
                    self.token_stack.pop()

            if not self.token_stack:
                # TODO actually give up earlier than this with the first
                # mismatch.
                raise ECMASyntaxError(
                    "Mismatched '%s' at %d:%d" % (
                        self.cur_token.value,
                        self.cur_token.lineno,
                        self.cur_token.colno,
                    )
                )

        # insert semicolon before restricted tokens
        # See section 7.9.1 ECMA262
        if (self.cur_token is not None
            and self.cur_token.type == 'LINE_TERMINATOR'
            and self.prev_token is not None
            and self.prev_token.type in ['BREAK', 'CONTINUE',
                                         'RETURN', 'THROW']):
            return self._create_semi_token(self.cur_token)

        return self.cur_token

    def _get_colno(self, token):
        # have a 1 offset to map nicer to commonly used/configured
        # text editors.
        return self._get_colno_lexpos(token.lexpos)

    def _get_colno_lexpos(self, lexpos):
        return lexpos - self.last_newline_lexpos + 1

    def lookup_colno(self, lineno, lexpos):
        """
        Look up a colno from the lineno and lexpos.
        """

        # have a 1 offset to map nicer to commonly used/configured
        # text editors.
        return lexpos - self.newline_idx[lineno - 1] + 1

    def _create_semi_token(self, orig_token):
        token = AutoLexToken()
        token.type = 'AUTOSEMI'
        token.value = ';'
        if orig_token is not None:
            token.lineno = orig_token.lineno
            # TODO figure out whether/how to normalize this with the
            # actual length of the original token...
            # Though, if actual use case boils down to error reporting,
            # line number is sufficient, and leaving it as 0 means it
            # shouldn't get dealt with during source map generation.
            token.colno = 0
            token.lexpos = orig_token.lexpos
        else:
            token.lineno = 0
            token.lexpos = 0
            token.colno = 0
        return token

    # iterator protocol
    def __iter__(self):
        return self

    def next(self):
        token = self.token()
        if not token:
            raise StopIteration

        return token

    __next__ = next

    states = (
        ('regex', 'exclusive'),
    )

    keywords = (
        'BREAK', 'CASE', 'CATCH', 'CONTINUE', 'DEBUGGER', 'DEFAULT', 'DELETE',
        'DO', 'ELSE', 'FINALLY', 'FOR', 'FUNCTION', 'IF', 'IN',
        'INSTANCEOF', 'NEW', 'RETURN', 'SWITCH', 'THIS', 'THROW', 'TRY',
        'TYPEOF', 'VAR', 'VOID', 'WHILE', 'WITH', 'NULL', 'TRUE', 'FALSE',
        # future reserved words - well, it's uncommented now to make
        # IE8 happy because it chokes up on minification:
        # obj["class"] -> obj.class
        'CLASS', 'CONST', 'ENUM', 'EXPORT', 'EXTENDS', 'IMPORT', 'SUPER',
    )
    keywords_dict = dict((key.lower(), key) for key in keywords)

    tokens = (
        # Punctuators
        'PERIOD', 'COMMA', 'SEMI', 'COLON',     # . , ; :
        'AUTOSEMI',                             # autogenerated ;
        'PLUS', 'MINUS', 'MULT', 'DIV', 'MOD',  # + - * / %
        'BAND', 'BOR', 'BXOR', 'BNOT',          # & | ^ ~
        'CONDOP',                               # conditional operator ?
        'NOT',                                  # !
        'LPAREN', 'RPAREN',                     # ( and )
        'LBRACE', 'RBRACE',                     # { and }
        'LBRACKET', 'RBRACKET',                 # [ and ]
        'EQ', 'EQEQ', 'NE',                     # = == !=
        'STREQ', 'STRNEQ',                      # === and !==
        'LT', 'GT',                             # < and >
        'LE', 'GE',                             # <= and >=
        'OR', 'AND',                            # || and &&
        'PLUSPLUS', 'MINUSMINUS',               # ++ and --
        'LSHIFT',                               # <<
        'RSHIFT', 'URSHIFT',                    # >> and >>>
        'PLUSEQUAL', 'MINUSEQUAL',              # += and -=
        'MULTEQUAL', 'DIVEQUAL',                # *= and /=
        'LSHIFTEQUAL',                          # <<=
        'RSHIFTEQUAL', 'URSHIFTEQUAL',          # >>= and >>>=
        'ANDEQUAL', 'MODEQUAL',                 # &= and %=
        'XOREQUAL', 'OREQUAL',                  # ^= and |=

        # Terminal types
        'NUMBER', 'STRING', 'ID', 'REGEX',

        # Properties
        'GETPROP', 'SETPROP',

        # Comments
        'LINE_COMMENT', 'BLOCK_COMMENT',

        'LINE_TERMINATOR',
    ) + keywords

    # adapted from https://bitbucket.org/ned/jslex
    t_regex_REGEX = r"""(?:
        /                       # opening slash
        # First character is..
        (?: [^*\\/[]            # anything but * \ / or [
        |   \\.                 # or an escape sequence
        |   \[                  # or a class, which has
                (?: [^\]\\]     # anything but \ or ]
                |   \\.         # or an escape sequence
                )*              # many times
            \]
        )
        # Following characters are same, except for excluding a star
        (?: [^\\/[]             # anything but \ / or [
        |   \\.                 # or an escape sequence
        |   \[                  # or a class, which has
                (?: [^\]\\]     # anything but \ or ]
                |   \\.         # or an escape sequence
                )*              # many times
            \]
        )*                      # many times
        /                       # closing slash
        [a-zA-Z0-9]*            # trailing flags
        )
        """

    t_regex_ignore = ' \t'

    def t_regex_error(self, token):
        raise ECMARegexSyntaxError(
            "Error parsing regular expression '%s' at %s:%s" % (
                token.value, token.lineno, self._get_colno(token))
        )

    # Punctuators
    t_PERIOD        = r'\.'
    t_COMMA         = r','
    t_SEMI          = r';'
    t_COLON         = r':'
    t_PLUS          = r'\+'
    t_MINUS         = r'-'
    t_MULT          = r'\*'
    t_DIV           = r'/'
    t_MOD           = r'%'
    t_BAND          = r'&'
    t_BOR           = r'\|'
    t_BXOR          = r'\^'
    t_BNOT          = r'~'
    t_CONDOP        = r'\?'
    t_NOT           = r'!'
    t_LPAREN        = r'\('
    t_RPAREN        = r'\)'
    t_LBRACE        = r'{'
    t_RBRACE        = r'}'
    t_LBRACKET      = r'\['
    t_RBRACKET      = r'\]'
    t_EQ            = r'='
    t_EQEQ          = r'=='
    t_NE            = r'!='
    t_STREQ         = r'==='
    t_STRNEQ        = r'!=='
    t_LT            = r'<'
    t_GT            = r'>'
    t_LE            = r'<='
    t_GE            = r'>='
    t_OR            = r'\|\|'
    t_AND           = r'&&'
    t_PLUSPLUS      = r'\+\+'
    t_MINUSMINUS    = r'--'
    t_LSHIFT        = r'<<'
    t_RSHIFT        = r'>>'
    t_URSHIFT       = r'>>>'
    t_PLUSEQUAL     = r'\+='
    t_MINUSEQUAL    = r'-='
    t_MULTEQUAL     = r'\*='
    t_DIVEQUAL      = r'/='
    t_LSHIFTEQUAL   = r'<<='
    t_RSHIFTEQUAL   = r'>>='
    t_URSHIFTEQUAL  = r'>>>='
    t_ANDEQUAL      = r'&='
    t_MODEQUAL      = r'%='
    t_XOREQUAL      = r'\^='
    t_OREQUAL       = r'\|='

    t_LINE_COMMENT  = r'//[^\r\n\u2028\u2029]*'
    t_BLOCK_COMMENT = r'/\*[^*]*\*+([^/*][^*]*\*+)*/'

    # 7.3 Line Terminators
    t_LINE_TERMINATOR = r'(\n|\r(?!\n)|\u2028|\u2029|\r\n)'

    t_ignore = (
        # space, tab, line tab, form feed, nbsp
        u' \t\x0b\x0c\xa0'
        # ogham space mark
        u'\u1680'
        # en quad .. hair space
        u'\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200A'
        # line sep, paragraph sep, narrow nbsp, med math, ideographic space
        u'\u2028\u2029\u202F\u205F\u3000'
        # unicode bom
        u'\uFEFF'
    )

    t_NUMBER = r"""
    (?:
        0[xX][0-9a-fA-F]+              # hex_integer_literal
     |  0[0-7]+                        # or octal_integer_literal
     |  (?:                            # or decimal_literal
            (?:0|[1-9][0-9]*)          # decimal_integer_literal
            \.                         # dot
            [0-9]*                     # decimal_digits_opt
            (?:[eE][+-]?[0-9]+)?       # exponent_part_opt
         |
            \.                         # dot
            [0-9]+                     # decimal_digits
            (?:[eE][+-]?[0-9]+)?       # exponent_part_opt
         |
            (?:0|[1-9][0-9]*)          # decimal_integer_literal
            (?:[eE][+-]?[0-9]+)?       # exponent_part_opt
         )
    )
    """

    string = r"""
    (?:
        # double quoted string
        (?:"                               # opening double quote
            (?: [^"\\\n\r\u2028\u2029]     # not ", \, line terminators; allow
                | \\(\n|\r(?!\n)|\u2028|\u2029|\r\n)  # line continuation
                | \\[a-tvwyzA-TVWYZ!-\/:-@\[-`{-~] # escaped chars
                | \\x[0-9a-fA-F]{2}        # hex_escape_sequence
                | \\u[0-9a-fA-F]{4}        # unicode_escape_sequence
                | \\(?:[1-7][0-7]{0,2}|[0-7]{2,3})  # octal_escape_sequence
                | \\0                      # <NUL> (15.10.2.11)
            )*?                            # zero or many times
        ")                                 # must have closing double quote
        |
        # single quoted string
        (?:'                               # opening single quote
            (?: [^'\\\n\r\u2028\u2029]     # not ', \, line terminators; allow
                | \\(\n|\r(?!\n)|\u2028|\u2029|\r\n)  # line continuation
                | \\[a-tvwyzA-TVWYZ!-\/:-@\[-`{-~] # escaped chars
                | \\x[0-9a-fA-F]{2}        # hex_escape_sequence
                | \\u[0-9a-fA-F]{4}        # unicode_escape_sequence
                | \\(?:[1-7][0-7]{0,2}|[0-7]{2,3}) # octal_escape_sequence
                | \\0                      # <NUL> (15.10.2.11)
            )*?                            # zero or many times
        ')                                 # must have closing single quote
    )
    """

    @ply.lex.TOKEN(string)
    def t_STRING(self, token):
        return token

    # XXX: <ZWNJ> <ZWJ> ?
    identifier_start = r'(?:' + r'[a-zA-Z_$]' + r'|' + LETTER + r')+'
    identifier_part = (
        r'(?:' + COMBINING_MARK + r'|' + r'[0-9a-zA-Z_$]' + r'|' + DIGIT +
        r'|' + CONNECTOR_PUNCTUATION + r')*'
    )
    identifier = identifier_start + identifier_part

    getprop = r'get' + r'(?=\s' + identifier + r')'

    @ply.lex.TOKEN(getprop)
    def t_GETPROP(self, token):
        return token

    setprop = r'set' + r'(?=\s' + identifier + r')'

    @ply.lex.TOKEN(setprop)
    def t_SETPROP(self, token):
        return token

    @ply.lex.TOKEN(identifier)
    def t_ID(self, token):
        token.type = self.keywords_dict.get(token.value, 'ID')
        return token

    def t_error(self, token):
        for handler in self.error_token_handlers:
            handler(self, token)

        if self.cur_token:
            # TODO make use of the extended calling signature when done
            raise ECMASyntaxError(
                'Illegal character %s at %s:%s after %s' % (
                    repr_compat(token.value[0]), token.lineno,
                    self._get_colno(token), format_lex_token(self.cur_token),
                )
            )
        else:
            raise ECMASyntaxError(
                'Illegal character %s at %s:%s' % (
                    repr_compat(token.value[0]), token.lineno,
                    self._get_colno(token),
                )
            )
